# encoding: utf-8
import tkinter as tk
import tkinter.ttk as ttk

from pygubu.widgets.tkscrollbarhelper import ScrollbarHelperBase


class TTKScrollbarHelperFactory(type):
    def __new__(cls, clsname, superclasses, attrs):
        return type.__new__(cls, str(clsname), superclasses, attrs)


ScrollbarHelper = TTKScrollbarHelperFactory(
    "ScrollbarHelper",
    (ScrollbarHelperBase, ttk.Frame, object),
    {"_framecls": ttk.Frame, "_sbarcls": ttk.Scrollbar},
)



from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.i18n import _
from ..customtkinter import _designer_tab_label, _plugin_uid
from .ctkbase import CTkBaseMixin, GCONTAINER

from customtkinter import CTkTabview


class CTkTabviewBO(CTkBaseMixin, BuilderObject):
    class_ = CTkTabview
    allow_bindings = False
    container = True
    properties = (
        "width",
        "height",
        "corner_radius",
        "border_width",
        "bg_color",
        "fg_color",
        "border_color",
        "segmented_button_fg_color",
        "segmented_button_selected_color",
        "segmented_button_selected_hover_color",
        "segmented_button_unselected_color",
        "segmented_button_unselected_hover_color",
        "text_color",
        "text_color_disabled",
        "command",
        "anchor",
        "state",
    )
    command_properties = ("command",)


_builder_uid = f"{_plugin_uid}.CTkTabview"
register_widget(
    _builder_uid,
    CTkTabviewBO,
    "CTkTabview",
    ("ttk", _designer_tab_label),
    group=GCONTAINER,
)


class CTkTabviewTabBO(BuilderObject):
    class_ = None
    container = True
    container_layout = True
    layout_required = False
    allow_bindings = False
    allowed_parents = (f"{_plugin_uid}.CTkTabview",)
    properties = ("label",)

    def _get_tab_name(self):
        return self.wmeta.properties.get("label", self.wmeta.identifier)

    def realize(self, parent, extra_init_args: dict = None):
        view = parent.get_child_master()
        self.widget = view.add(self._get_tab_name())
        return self.widget

    def configure(self, target=None):
        pass

    #
    # Code generation methods
    #
    def code_realize(self, boparent, code_identifier=None):
        view = boparent.code_child_master()
        tabid = self.code_identifier()
        tab_name = self._get_tab_name()
        lines = [f'{tabid} = {view}.add("{tab_name}")']
        return lines

    def code_configure(self, targetid=None):
        return tuple()


_builder_uid = f"{_plugin_uid}.CTkTabview.Tab"

CTkTabviewBO.add_allowed_child(_builder_uid)

register_widget(
    _builder_uid,
    CTkTabviewTabBO,
    "CTkTabview.Tab",
    ("ttk", _designer_tab_label),
    group=GCONTAINER,
)

_help = _("The 'name' argument of method: CTkTabview.add(self, name: str)")
register_custom_property(_builder_uid, "label", "entry", help=_help)



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk

import fixpath
import pygubu
import support


class TestCommandIdArg(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_command_id_arg.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("mainwindow")
        self.menu = builder.get_object("menu1")

    def tearDown(self):
        support.root_withdraw()

    def test_idtocommand(self):
        success = []
        received_id = []

        class AnObject:
            def menu_item_clicked(self, itemid):
                success.append(1)
                received_id.append(itemid)

        cbobj = AnObject()
        self.builder.connect_callbacks(cbobj)
        self.menu.invoke(0)
        self.widget.update()

        self.assertTrue(success)
        self.assertEqual(received_id[0], "menu_item_1")
        self.widget.destroy()



__all__ = ["remove_binding", "ApplicationLevelBindManager"]

import logging
import platform
import tkinter as tk

from pygubu.utils.widget import iter_to_toplevel


logger = logging.getLogger(__name__)


def bindings(widget, seq):
    return [x for x in widget.bind(seq).splitlines() if x.strip()]


def _funcid(binding):
    return binding.split()[1][3:]


def remove_binding(widget, seq, index=None, funcid=None):
    b = bindings(widget, seq)

    if index is not None:
        try:
            binding = b[index]
            widget.unbind(seq, _funcid(binding))
            b.remove(binding)
        except IndexError:
            logger.info("Binding #%d not defined.", index)
            return

    elif funcid:
        binding = None
        for x in b:
            if _funcid(x) == funcid:
                binding = x
                b.remove(binding)
                widget.unbind(seq, funcid)
                break
        if not binding:
            logger.info('Binding "%s" not defined.', funcid)
            return
    else:
        raise ValueError("No index or function id defined.")

    for x in b:
        widget.bind(seq, "+" + x, 1)


class AppBindManagerBase(object):
    # Acces to tk instance
    master: tk.Widget = None
    # Mouse wheel support
    mw_listeners = []  # Mousewheel listeners
    mw_initialized = False

    @classmethod
    def on_mousewheel(cls, event):
        widget_below = event.widget
        if not isinstance(widget_below, tk.Widget):
            try:
                widget_below = cls.master.winfo_containing(
                    event.x_root, event.y_root
                )
            except KeyError:
                widget_below = None
        if widget_below:
            for w in iter_to_toplevel(widget_below):
                if w in cls.mw_listeners:
                    can_keep_scrolling = w.on_mousewheel(event)
                    if can_keep_scrolling:
                        break

    @classmethod
    def mousewheel_bind(cls, widget):
        if widget not in cls.mw_listeners:
            cls.mw_listeners.append(widget)

    @classmethod
    def mousewheel_unbind(cls, widget):
        if widget in cls.mw_listeners:
            cls.mw_listeners.remove(widget)

    @classmethod
    def init_mousewheel_binding(cls, master):
        if not cls.mw_initialized:
            cls.master = master.winfo_toplevel()
            _os = platform.system()
            if _os in ("Linux", "OpenBSD", "FreeBSD"):
                master.bind_all("<4>", cls.on_mousewheel, add="+")
                master.bind_all("<5>", cls.on_mousewheel, add="+")
            else:
                # Windows and MacOS
                master.bind_all(
                    "<MouseWheel>",
                    cls.on_mousewheel,
                    add="+",
                )
            cls.mw_initialized = True

    @classmethod
    def make_onmousewheel_cb(cls, widget, orient, factor=1):
        """Create a callback to manage mousewheel events

        orient: string (posible values: ('x', 'y'))
        widget: widget that implement tk xview and yview methods
        """
        _os = platform.system()
        view_command = getattr(widget, orient + "view")
        if _os in ("Linux", "OpenBSD", "FreeBSD"):

            def on_mousewheel(event):
                can_keep_scrolling = True
                if event.num == 4:
                    view_command("scroll", (-1) * factor, "units")
                    scroll_rs = view_command()
                    can_keep_scrolling = scroll_rs[0] != 0.0
                elif event.num == 5:
                    view_command("scroll", factor, "units")
                    scroll_rs = view_command()
                    can_keep_scrolling = scroll_rs[1] != 1.0
                return can_keep_scrolling

        elif _os == "Windows":

            def on_mousewheel(event):
                view_command(
                    "scroll", (-1) * int((event.delta / 120) * factor), "units"
                )
                scroll_rs = view_command()
                can_keep_scrolling = scroll_rs[0] != 0.0
                if event.delta < 0:
                    can_keep_scrolling = scroll_rs[1] != 1.0
                return can_keep_scrolling

        elif _os == "Darwin":

            def on_mousewheel(event):
                view_command("scroll", event.delta, "units")
                scroll_rs = view_command()
                can_keep_scrolling = scroll_rs[0] != 0.0
                if event.delta < 0:
                    can_keep_scrolling = scroll_rs[1] != 1.0
                return can_keep_scrolling

        else:
            # FIXME: unknown platform scroll method
            def on_mousewheel(event):
                pass

        return on_mousewheel


class ApplicationLevelBindManager(AppBindManagerBase):
    ...



class FieldWidget:
    def __init__(self, *args, field_name: str, **kw):
        super().__init__(*args, **kw)
        self.field_name = field_name

    def wset_value(self, value):
        # will set the value in the widget format
        raise NotImplementedError(
            f"Subclasses must define this method. {self.__class__}"
            f" for field class {self.__class__}"
        )

    def wget_value(self):
        # Get value in the widget
        raise NotImplementedError(
            f"Subclasses must define this method. {self.__class__}"
            f" for field class {self.__class__}"
        )

    def wmark_invalid(self, state: bool):
        # Visually mark the widget as invalid depending on state parameter.
        raise NotImplementedError(
            f"Subclasses must define this method. {self.__class__}"
            f" for field class {self.__class__}"
        )

    def wis_disabled(self) -> bool:
        raise NotImplementedError(
            f"Subclasses must define this method. {self.__class__}"
            f" for field class {self.__class__}"
        )


class WidgetInfoBase:
    def __init__(self, *args, field_name: str, **kw):
        self.field_name = field_name
        super().__init__(*args, **kw)

    def show_error(self, error):
        raise NotImplementedError

    def show_help(self, message):
        raise NotImplementedError

    def clear(self):
        raise NotImplementedError


class WidgetInfo(WidgetInfoBase):
    pass


class FormInfo(WidgetInfoBase):
    pass



import sys
import os
import tkinter as tk
import tkinter.ttk as ttk
import math
import timeit
import logging

from tkinter import font
from dataclasses import dataclass
from ..base import IThemeBuilder
from ..color import ColorUtil
from .assets import AssetCreator
from .config import (
    FIX_LINUX_FILE_DIALOG,
    PRIMARY,
    SECONDARY,
    SUCCESS,
    INFO,
    WARNING,
    DANGER,
    LIGHT,
    DARK,
    TTK_CLAM,
    TTK_ALT,
    TTK_DEFAULT,
)

logger = logging.getLogger(__name__)


@dataclass
class Colors:
    primary: str
    secondary: str
    success: str
    info: str
    warning: str
    danger: str
    light: str
    dark: str
    # --
    bg: str
    fg: str
    selectbg: str
    selectfg: str
    border: str
    inputfg: str
    inputbg: str
    active: str

    def names(self):
        return iter(
            [
                PRIMARY,
                SECONDARY,
                SUCCESS,
                INFO,
                WARNING,
                DANGER,
                LIGHT,
                DARK,
            ]
        )

    def get_foreground(self, color_label):
        if color_label == LIGHT:
            return self.dark
        elif color_label == DARK:
            return self.light
        else:
            return self.selectfg

    def get_color(self, name):
        return getattr(self, name)


class ThemeDefinition:
    THEME_LIGHT = LIGHT
    THEME_DARK = DARK

    def __init__(self, name: str, colors: dict, theme_type=THEME_LIGHT):
        self.name = name
        self.colors = Colors(**colors)
        self.type = theme_type


_colect_styles = True if os.getenv("PYGUBU_DESIGNER_RUNNING") else False


class BootstrapThemeBuilder(IThemeBuilder):
    generated_styles = set()

    def __init__(self, theme_def):
        self.theme = theme_def
        self.tk_master = None
        self.colorutil = None
        self.existing_elements = {}
        self.assets = AssetCreator(self)
        self.tk_widgets_options = {}

    def theme_name(self):
        return self.theme.name

    def theme_parent(self):
        return "clam"

    def theme_settings(self):
        return self.build_theme_settings()

    def tk_palette(self):
        return self.build_tk_palette()

    def db_settings(self):
        self.build_tk_widget_settings()
        db = {}
        for widget, options in self.tk_widgets_options.items():
            pattern = f"*{widget}."
            db[pattern] = options.copy()
        return db

    def create(self, master):
        self.tk_master = master
        self.colorutil = ColorUtil(master)
        super().create(master)

        if sys.platform == "linux" and FIX_LINUX_FILE_DIALOG:
            self._install_linux_filedialog_hack(master)

    def apply(self, master: tk.Widget):
        super().apply(master)

        if sys.platform == "linux" and FIX_LINUX_FILE_DIALOG:
            self._update_linux_filedialog_bg(master)

    def _get_filedialog_bg_color(self):
        color = self.colors.bg if self.is_light_theme else self.colors.selectbg
        color = self.colorutil.update_hsv(color, vd=0.18)
        return color

    def _install_linux_filedialog_hack(self, master):
        """Install script to modify dialog background color.
        For now, this only works for the first time the dialog window is created.
        """
        script = """
# Enable filedialog namespace
catch {tk_getOpenFile -badoption}

# Install hack function
if { [info exists ::tk::dialog::file::pbs_hack ] == 0 } {
  set ::tk::dialog::file::pbs_hack 1
  # Set color for background
  set ::tk::dialog::file::pbs_filedialog_bg {color}

  rename ::tk::dialog::file::Create ::tk::dialog::file::_pbs_filedialog_create
  proc ::tk::dialog::file::Create {w class} {
    eval ::tk::dialog::file::_pbs_filedialog_create $w $class
    $w.contents.icons.cHull.canvas configure -background $::tk::dialog::file::pbs_filedialog_bg
  }
}
"""
        script = script.replace("{color}", self._get_filedialog_bg_color())
        master.tk.eval(script)

    def _update_linux_filedialog_bg(self, master):
        script = """
if { [info exists ::tk::dialog::file::pbs_hack ] == 1 } {
  set ::tk::dialog::file::pbs_filedialog_bg {color}
}
"""
        script = script.replace("{color}", self._get_filedialog_bg_color())
        master.tk.eval(script)

    def update_current_widgets(self, master):
        if not isinstance(master, tk.Tk):
            master = master.winfo_toplevel()
        style = ttk.Style(self.tk_master)
        self._walk_and_update_widget(master, style)

    def _walk_and_update_widget(self, widget, style, level=0):
        if isinstance(widget, tk.Misc):
            wclass = widget.winfo_class()
            # wstyle = None
            if isinstance(widget, ttk.Widget) and wclass == "TCombobox":
                # wstyle = widget.cget("style")
                # curr_style = wstyle if wstyle else wclass
                self.update_combobox_popdown_style(widget)
            else:
                if wclass in self.tk_widgets_options:
                    # some widget may block the property (like customtkinter)
                    # so add try here.
                    try:
                        widget.configure(**self.tk_widgets_options[wclass])
                    except Exception as e:
                        logger.debug(e)
            # print("  " * level, f"{wclass} {wstyle}")
            for k, v in widget.children.items():
                self._walk_and_update_widget(v, style, level + 1)

    @property
    def colors(self):
        return self.theme.colors

    @property
    def is_light_theme(self):
        return self.theme.type == ThemeDefinition.THEME_LIGHT

    def scale_size(self, size):
        """Scale the size of images and other assets based on the
        scaling factor of ttk to ensure that the image matches the
        screen resolution.

        Parameters:

            size (Union[int, List, Tuple]):
                A single integer or an iterable of integers
        """
        winsys = self.tk_master.tk.call("tk", "windowingsystem")
        if winsys == "aqua":
            BASELINE = 1.000492368291482
        else:
            BASELINE = 1.33398982438864281
        scaling = self.tk_master.tk.call("tk", "scaling")
        factor = scaling / BASELINE

        if isinstance(size, int) or isinstance(size, float):
            return math.ceil(size * factor)
        elif isinstance(size, tuple) or isinstance(size, list):
            return [math.ceil(x * factor) for x in size]

    def build_tk_widget_settings(self):
        # widget.option_add('*Text*Font', 'TkDefaultFont')
        wo = {
            "Tk": {"background": self.colors.bg},
            "Toplevel": {"background": self.colors.bg},
            "Canvas": {
                "background": self.colors.bg,
                "highlightthickness": 0,
            },
            "Label": {
                "foreground": self.colors.fg,
                "background": self.colors.bg,
            },
            "Frame": {"background": self.colors.bg},
            "Checkbutton": {
                "activebackground": self.colors.bg,
                "activeforeground": self.colors.primary,
                "background": self.colors.bg,
                "foreground": self.colors.fg,
                "selectcolor": self.colors.bg,
            },
            "Radiobutton": {
                "activebackground": self.colors.bg,
                "activeforeground": self.colors.primary,
                "background": self.colors.bg,
                "foreground": self.colors.fg,
                "selectcolor": self.colors.bg,
            },
            "Menu": {
                "tearoff": False,
                "activebackground": self.colors.selectbg,
                "activeforeground": self.colors.selectfg,
                "foreground": self.colors.fg,
                "selectcolor": self.colors.primary,
                "background": self.colors.bg,
                "relief": tk.FLAT,
                "borderwidth": 0,
            },
        }
        # buttons
        background = self.colors.primary
        foreground = self.colors.selectfg
        activebackground = self.colorutil.update_hsv(
            self.colors.primary, vd=-0.1
        )
        wo["Button"] = {
            "background": background,
            "foreground": foreground,
            "relief": tk.FLAT,
            "borderwidth": 0,
            "activebackground": activebackground,
            "highlightbackground": self.colors.selectfg,
        }
        # Entry
        bordercolor = (
            self.colors.border if self.is_light_theme else self.colors.selectbg
        )
        wo["Entry"] = {
            "relief": tk.FLAT,
            "highlightthickness": 1,
            "foreground": self.colors.inputfg,
            "highlightbackground": bordercolor,
            "highlightcolor": self.colors.primary,
            "background": self.colors.inputbg,
            "insertbackground": self.colors.inputfg,
            "insertwidth": 1,
        }
        # Scale
        bordercolor = (
            self.colors.border if self.is_light_theme else self.colors.selectbg
        )
        activecolor = self.colorutil.update_hsv(self.colors.primary, vd=-0.2)
        wo["Scale"] = {
            "background": self.colors.primary,
            "showvalue": False,
            "sliderrelief": tk.FLAT,
            "borderwidth": 0,
            "activebackground": activecolor,
            "highlightthickness": 1,
            "highlightcolor": bordercolor,
            "highlightbackground": bordercolor,
            "troughcolor": self.colors.inputbg,
        }
        # Spinbox
        bordercolor = (
            self.colors.border if self.is_light_theme else self.colors.selectbg
        )
        wo["Spinbox"] = {
            "relief": tk.FLAT,
            "highlightthickness": 1,
            "foreground": self.colors.inputfg,
            "highlightbackground": bordercolor,
            "highlightcolor": self.colors.primary,
            "background": self.colors.inputbg,
            "buttonbackground": self.colors.inputbg,
            "insertbackground": self.colors.inputfg,
            "insertwidth": 1,
            # these options should work, but do not have any affect
            "buttonuprelief": tk.FLAT,
            "buttondownrelief": tk.SUNKEN,
        }
        # Listbox
        bordercolor = (
            self.colors.border if self.is_light_theme else self.colors.selectbg
        )
        wo["Spinbox"] = {
            "foreground": self.colors.inputfg,
            "background": self.colors.inputbg,
            "selectbackground": self.colors.selectbg,
            "selectforeground": self.colors.selectfg,
            "highlightcolor": self.colors.primary,
            "highlightbackground": bordercolor,
            "highlightthickness": 1,
            "activestyle": "none",
            "relief": tk.FLAT,
        }
        # Menubutton
        activebackground = self.colorutil.update_hsv(
            self.colors.primary, vd=-0.2
        )
        wo["Menubutton"] = {
            "background": self.colors.primary,
            "foreground": self.colors.selectfg,
            "activebackground": activebackground,
            "activeforeground": self.colors.selectfg,
            "borderwidth": 0,
        }
        # LabelFrame
        bordercolor = (
            self.colors.border if self.is_light_theme else self.colors.selectbg
        )
        wo["LabelFrame"] = {
            "highlightcolor": bordercolor,
            "foreground": self.colors.fg,
            "borderwidth": 1,
            "highlightthickness": 0,
            "background": self.colors.bg,
        }
        # Text
        bordercolor = (
            self.colors.border if self.is_light_theme else self.colors.selectbg
        )
        focuscolor = bordercolor
        wo["Text"] = {
            "background": self.colors.inputbg,
            "foreground": self.colors.inputfg,
            "highlightcolor": focuscolor,
            "highlightbackground": bordercolor,
            "insertbackground": self.colors.inputfg,
            "selectbackground": self.colors.selectbg,
            "selectforeground": self.colors.selectfg,
            "insertwidth": 1,
            "highlightthickness": 1,
            "relief": tk.FLAT,
            "padx": 5,
            "pady": 5,
        }
        self.tk_widgets_options = wo

    def build_theme_settings(self):
        tstart = timeit.default_timer()
        settings = {}
        self.create_default_style(settings)

        colornames = [c for c in self.colors.names()]
        colornames.insert(0, None)
        for name in colornames:
            # Frames
            self.create_frame_style(settings, name)
            self.create_labelframe_style(settings, name)
            # Labels
            self.create_label_style(settings, name)
            self.create_inverse_label_style(settings, name)
            # Buttons
            self.create_button_style(settings, name)
            self.create_toolbutton_style(settings, name)
            self.create_outline_menubutton_style(settings, name)
            self.create_outline_toolbutton_style(settings, name)
            self.create_link_button_style(settings, name)
            self.create_radiobutton_style(settings, name)
            self.create_menubutton_style(settings, name)
            self.create_checkbutton_style(settings, name)
            self.create_square_toggle_style(settings, name)
            self.create_round_toggle_style(settings, name)
            # Entry
            self.create_entry_style(settings, name)
            # Combo
            self.create_combobox_style(settings, name)
            # Panedwindow
            self.create_panedwindow_style(settings, name)
            # Notebook
            self.create_notebook_style(settings, name)
            # separator
            self.create_separator_style(settings, name)
            # Progress bars
            self.create_progressbar_style(settings, name)
            self.create_striped_progressbar_style(settings, name)
            # Scale
            self.create_scale_style(settings, name)
            # scrollbar
            self.create_scrollbar_style(settings, name)
            self.create_round_scrollbar_style(settings, name)
            # spinbox
            self.create_spinbox_style(settings, name)
            # treeview
            self.create_treeview_style(settings, name)
            self.create_table_treeview_style(settings, name)
            # sizegrip
            self.create_sizegrip_style(settings, name)

        tfinish = timeit.default_timer() - tstart
        logger.debug("Bootstrap ttk theme build time: %s", tfinish)
        return settings

    def build_tk_palette(self):
        return {
            "background": self.colors.bg,
            "foreground": self.colors.fg,
            "highlightColor": self.colors.primary,
            "selectBackground": self.colors.selectbg,
            "selectForeground": self.colors.selectfg,
            "activeBackground": self.colors.selectbg,
            "activeForeground": self.colors.selectfg,
        }

    def _register_style(self, style_name):
        if _colect_styles:
            self.generated_styles.add(style_name)

    def update_combobox_popdown_style(self, widget):
        """Update the legacy ttk.Combobox elements. This method is
        called every time the theme is changed in order to ensure
        that the legacy tkinter components embedded in this ttk widget
        are styled appropriate to the current theme.

        The ttk.Combobox contains several elements that are not styled
        using the ttk theme engine. This includes the **popdownwindow**
        and the **scrollbar**. Both of these widgets are configured
        manually using calls to tcl/tk.

        Parameters:

            widget (ttk.Combobox):
                The combobox element to be updated.
        """
        if self.is_light_theme:
            bordercolor = self.colors.border
        else:
            bordercolor = self.colors.selectbg

        tk_settings = []
        tk_settings.extend(["-borderwidth", 2])
        tk_settings.extend(["-highlightthickness", 1])
        tk_settings.extend(["-highlightcolor", bordercolor])
        tk_settings.extend(["-background", self.colors.inputbg])
        tk_settings.extend(["-foreground", self.colors.inputfg])
        tk_settings.extend(["-selectbackground", self.colors.selectbg])
        tk_settings.extend(["-selectforeground", self.colors.selectfg])

        # set popdown style
        popdown = widget.tk.eval(f"ttk::combobox::PopdownWindow {widget}")
        widget.tk.call(f"{popdown}.f.l", "configure", *tk_settings)

        # set scrollbar style
        sb_style = "TCombobox.Vertical.TScrollbar"
        widget.tk.call(f"{popdown}.f.sb", "configure", "-style", sb_style)

    def create_default_style(self, settings: dict):
        settings["."] = {
            "configure": {
                "background": self.colors.bg,
                "darkcolor": self.colors.border,
                "foreground": self.colors.fg,
                "troughcolor": self.colors.bg,
                "selectbg": self.colors.selectbg,
                "selectfg": self.colors.selectfg,
                "selectbackground": self.colors.selectbg,
                "selectforeground": self.colors.selectfg,
                "fieldbackground": self.colors.bg,
                # "font": "TkDefaultFont",
                "borderwidth": 1,
                "focuscolor": "",
            },
        }

    def create_frame_style(self, settings, colorname=None):
        """Create a style for the ttk.Frame widget."""
        STYLE = "TFrame"

        if any([colorname is None, not colorname]):
            ttkstyle = STYLE
            background = self.colors.bg
        else:
            ttkstyle = f"{colorname}.{STYLE}"
            background = self.colors.get_color(colorname)

        settings[ttkstyle] = {
            "configure": {
                "background": background,
            }
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_labelframe_style(self, settings, colorname=None):
        """Create a style for the ttk.Labelframe widget."""
        STYLE = "TLabelframe"

        background = self.colors.bg

        if any([colorname is None, not colorname]):
            foreground = self.colors.fg
            ttkstyle = STYLE

            if self.is_light_theme:
                bordercolor = self.colors.border
            else:
                bordercolor = self.colors.selectbg

        else:
            foreground = self.colors.get_color(colorname)
            bordercolor = foreground
            ttkstyle = f"{colorname}.{STYLE}"

        settings[f"{ttkstyle}.Label"] = {
            "configure": {
                "foreground": foreground,
                "background": background,
            }
        }
        settings[ttkstyle] = {
            "configure": {
                "relief": tk.RAISED,
                "borderwidth": 1,
                "bordercolor": bordercolor,
                "lightcolor": background,
                "darkcolor": background,
                "background": background,
            }
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_label_style(self, settings, colorname=None):
        """Create a standard style for the ttk.Label widget.

        Parameters:

            colorname (str):
                The color label used to style the widget.
        """
        STYLE = "TLabel"

        if any([colorname is None, colorname == ""]):
            ttkstyle = STYLE
            foreground = self.colors.fg
            background = self.colors.bg
        else:
            ttkstyle = f"{colorname}.{STYLE}"
            foreground = self.colors.get_color(colorname)
            background = self.colors.bg

        # standard label
        settings[ttkstyle] = {
            "configure": {"foreground": foreground, "background": background}
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_inverse_label_style(self, settings, colorname=None):
        """Create an inverted style for the ttk.Label."""
        STYLE_INVERSE = "Inverse.TLabel"

        if any([colorname is None, not colorname]):
            ttkstyle = STYLE_INVERSE
            background = self.colors.fg
            foreground = self.colors.bg
        else:
            ttkstyle = f"{colorname}.{STYLE_INVERSE}"
            background = self.colors.get_color(colorname)
            foreground = self.colors.get_foreground(colorname)

        settings[ttkstyle] = {
            "configure": {"foreground": foreground, "background": background}
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_button_style(self, settings, colorname=None):
        STYLE = "TButton"

        if any([colorname is None, colorname == ""]):
            ttkstyle = STYLE
            foreground = self.colors.get_foreground(PRIMARY)
            background = self.colors.primary
        else:
            ttkstyle = f"{colorname}.{STYLE}"
            foreground = self.colors.get_foreground(colorname)
            background = self.colors.get_color(colorname)

        bordercolor = background
        disabled_bg = self.colorutil.make_transparent(
            0.10, self.colors.fg, self.colors.bg
        )
        disabled_fg = self.colorutil.make_transparent(
            0.30, self.colors.fg, self.colors.bg
        )
        pressed = self.colorutil.make_transparent(
            0.80, background, self.colors.bg
        )
        hover = self.colorutil.make_transparent(
            0.90, background, self.colors.bg
        )

        settings[ttkstyle] = {
            "configure": {
                "foreground": foreground,
                "background": background,
                "bordercolor": bordercolor,
                "darkcolor": background,
                "lightcolor": background,
                "relief": tk.RAISED,
                "focusthickness": 0,
                "focuscolor": foreground,
                "padding": (10, 5),
                "anchor": tk.CENTER,
            },
            "map": {
                "foreground": [("disabled", disabled_fg)],
                "background": [
                    ("disabled", disabled_bg),
                    ("pressed !disabled", pressed),
                    ("hover !disabled", hover),
                ],
                "bordercolor": [("disabled", disabled_bg)],
                "darkcolor": [
                    ("disabled", disabled_bg),
                    ("pressed !disabled", pressed),
                    ("hover !disabled", hover),
                ],
                "lightcolor": [
                    ("disabled", disabled_bg),
                    ("pressed !disabled", pressed),
                    ("hover !disabled", hover),
                ],
            },
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_toolbutton_style(self, settings, colorname=None):
        """Create a solid toolbutton style for the ttk.Checkbutton"""
        STYLE = "Toolbutton"

        if any([colorname is None, colorname == ""]):
            ttkstyle = STYLE
            toggle_on = self.colors.primary
        else:
            ttkstyle = f"{colorname}.{STYLE}"
            toggle_on = self.colors.get_color(colorname)

        foreground = self.colors.get_foreground(colorname)

        if self.is_light_theme:
            toggle_off = self.colors.border
        else:
            toggle_off = self.colors.selectbg

        disabled_bg = self.colorutil.make_transparent(
            0.10, self.colors.fg, self.colors.bg
        )
        disabled_fg = self.colorutil.make_transparent(
            0.30, self.colors.fg, self.colors.bg
        )

        settings[ttkstyle] = {
            "configure": {
                "foreground": self.colors.selectfg,
                "background": toggle_off,
                "bordercolor": toggle_off,
                "darkcolor": toggle_off,
                "lightcolor": toggle_off,
                "relief": tk.RAISED,
                "focusthickness": 0,
                "focuscolor": "",
                "padding": (10, 5),
                "anchor": tk.CENTER,
            },
            "map": {
                "foreground": [
                    ("disabled", disabled_fg),
                    ("hover", foreground),
                    ("selected", foreground),
                ],
                "background": [
                    ("disabled", disabled_bg),
                    ("pressed !disabled", toggle_on),
                    ("selected !disabled", toggle_on),
                    ("hover !disabled", toggle_on),
                ],
                "bordercolor": [
                    ("disabled", disabled_bg),
                    ("pressed !disabled", toggle_on),
                    ("selected !disabled", toggle_on),
                    ("hover !disabled", toggle_on),
                ],
                "darkcolor": [
                    ("disabled", disabled_bg),
                    ("pressed !disabled", toggle_on),
                    ("selected !disabled", toggle_on),
                    ("hover !disabled", toggle_on),
                ],
                "lightcolor": [
                    ("disabled", disabled_bg),
                    ("pressed !disabled", toggle_on),
                    ("selected !disabled", toggle_on),
                    ("hover !disabled", toggle_on),
                ],
            },
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_outline_menubutton_style(self, settings, colorname=None):
        """Create an outline button style for the ttk.Menubutton widget"""
        STYLE = "Outline.TMenubutton"

        disabled_fg = self.colorutil.make_transparent(
            0.30, self.colors.fg, self.colors.bg
        )

        if any([colorname is None, not colorname]):
            ttkstyle = STYLE
            colorname = PRIMARY
        else:
            ttkstyle = f"{colorname}.{STYLE}"

        foreground = self.colors.get_color(colorname)
        background = self.colors.get_foreground(colorname)
        foreground_pressed = background
        bordercolor = foreground
        pressed = foreground
        hover = foreground

        settings[ttkstyle] = {
            "configure": {
                "foreground": foreground,
                "background": self.colors.bg,
                "bordercolor": bordercolor,
                "darkcolor": self.colors.bg,
                "lightcolor": self.colors.bg,
                "relief": tk.RAISED,
                "focusthickness": 0,
                "focuscolor": foreground,
                "padding": (10, 5),
                "arrowcolor": foreground,
                "arrowpadding": (0, 0, 15, 0),
                "arrowsize": self.scale_size(4),
            },
            "map": {
                "foreground": [
                    ("disabled", disabled_fg),
                    ("pressed !disabled", foreground_pressed),
                    ("hover !disabled", foreground_pressed),
                ],
                "background": [
                    ("pressed !disabled", pressed),
                    ("hover !disabled", hover),
                ],
                "bordercolor": [
                    ("disabled", disabled_fg),
                    ("pressed", pressed),
                    ("hover", hover),
                ],
                "darkcolor": [
                    ("pressed !disabled", pressed),
                    ("hover !disabled", hover),
                ],
                "lightcolor": [
                    ("pressed !disabled", pressed),
                    ("hover !disabled", hover),
                ],
                "arrowcolor": [
                    ("disabled", disabled_fg),
                    ("pressed", foreground_pressed),
                    ("hover", foreground_pressed),
                ],
            },
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_outline_toolbutton_style(self, settings, colorname=None):
        """Create an outline toolbutton style for the ttk.Checkbutton
        and ttk.Radiobutton widgets.
        """
        STYLE = "Outline.Toolbutton"

        disabled_fg = self.colorutil.make_transparent(
            0.30, self.colors.fg, self.colors.bg
        )

        if any([colorname is None, colorname == ""]):
            ttkstyle = STYLE
            colorname = PRIMARY
        else:
            ttkstyle = f"{colorname}.{STYLE}"

        foreground = self.colors.get_color(colorname)
        background = self.colors.get_foreground(colorname)
        foreground_pressed = background
        bordercolor = foreground
        pressed = foreground
        hover = foreground

        settings[ttkstyle] = {
            "configure": {
                "foreground": foreground,
                "background": self.colors.bg,
                "bordercolor": bordercolor,
                "darkcolor": self.colors.bg,
                "lightcolor": self.colors.bg,
                "relief": tk.RAISED,
                "focusthickness": 0,
                "focuscolor": foreground,
                "padding": (10, 5),
                "anchor": tk.CENTER,
                "arrowcolor": foreground,
                "arrowpadding": (0, 0, 15, 0),
                "arrowsize": 3,
            },
            "map": {
                "foreground": [
                    ("disabled", disabled_fg),
                    ("pressed !disabled", foreground_pressed),
                    ("selected !disabled", foreground_pressed),
                    ("hover !disabled", foreground_pressed),
                ],
                "background": [
                    ("pressed !disabled", pressed),
                    ("selected !disabled", pressed),
                    ("hover !disabled", hover),
                ],
                "bordercolor": [
                    ("disabled", disabled_fg),
                    ("pressed !disabled", pressed),
                    ("selected !disabled", pressed),
                    ("hover !disabled", hover),
                ],
                "darkcolor": [
                    ("disabled", self.colors.bg),
                    ("pressed !disabled", pressed),
                    ("selected !disabled", pressed),
                    ("hover !disabled", hover),
                ],
                "lightcolor": [
                    ("disabled", self.colors.bg),
                    ("pressed !disabled", pressed),
                    ("selected !disabled", pressed),
                    ("hover !disabled", hover),
                ],
            },
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_link_button_style(self, settings, colorname=None):
        """Create a link button style for the ttk.Button widget."""
        STYLE = "Link.TButton"

        pressed = self.colors.info
        hover = self.colors.info

        if any([colorname is None, not colorname]):
            foreground = self.colors.fg
            ttkstyle = STYLE
        elif colorname == LIGHT:
            foreground = self.colors.fg
            ttkstyle = f"{colorname}.{STYLE}"
        else:
            foreground = self.colors.get_color(colorname)
            ttkstyle = f"{colorname}.{STYLE}"

        disabled_fg = self.colorutil.make_transparent(
            0.30, self.colors.fg, self.colors.bg
        )

        settings[ttkstyle] = {
            "configure": {
                "foreground": foreground,
                "background": self.colors.bg,
                "bordercolor": self.colors.bg,
                "darkcolor": self.colors.bg,
                "lightcolor": self.colors.bg,
                "relief": tk.RAISED,
                "focusthickness": 0,
                "focuscolor": foreground,
                "anchor": tk.CENTER,
                "padding": (10, 5),
            },
            "map": {
                "shiftrelief=": [("pressed !disabled", -1)],
                "foreground": [
                    ("disabled", disabled_fg),
                    ("pressed", "!disabled", pressed),
                    ("hover", "!disabled", hover),
                ],
                "focuscolor": [
                    ("pressed !disabled", pressed),
                    ("hover", "!disabled", pressed),
                ],
                "background": [
                    ("disabled", self.colors.bg),
                    ("pressed !disabled", self.colors.bg),
                    ("hover", "!disabled", self.colors.bg),
                ],
                "bordercolor": [
                    ("disabled", self.colors.bg),
                    ("pressed !disabled", self.colors.bg),
                    ("hover !disabled", self.colors.bg),
                ],
                "darkcolor": [
                    ("disabled", self.colors.bg),
                    ("pressed !disabled", self.colors.bg),
                    ("hover !disabled", self.colors.bg),
                ],
                "lightcolor": [
                    ("disabled", self.colors.bg),
                    ("pressed !disabled", self.colors.bg),
                    ("hover !disabled", self.colors.bg),
                ],
            },
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_entry_style(self, settings, colorname=None):
        """Create a style for the ttk.Entry widget."""
        STYLE = "TEntry"

        # general default colors
        if self.is_light_theme:
            disabled_fg = self.colors.border
            bordercolor = self.colors.border
            readonly = self.colors.light
        else:
            disabled_fg = self.colors.selectbg
            bordercolor = self.colors.selectbg
            readonly = bordercolor

        if any([colorname is None, not colorname]):
            # default style
            ttkstyle = STYLE
            focuscolor = self.colors.primary
        else:
            # colored style
            ttkstyle = f"{colorname}.{STYLE}"
            focuscolor = self.colors.get_color(colorname)
            bordercolor = focuscolor

        settings[ttkstyle] = {
            "configure": {
                "bordercolor": bordercolor,
                "darkcolor": self.colors.inputbg,
                "lightcolor": self.colors.inputbg,
                "fieldbackground": self.colors.inputbg,
                "foreground": self.colors.inputfg,
                "insertcolor": self.colors.inputfg,
                "padding": 5,
            },
            "map": {
                "foreground": [("disabled", disabled_fg)],
                "fieldbackground": [("readonly", readonly)],
                "bordercolor": [
                    ("invalid", self.colors.danger),
                    ("focus !disabled", focuscolor),
                    ("hover !disabled", focuscolor),
                ],
                "lightcolor": [
                    ("focus invalid", self.colors.danger),
                    ("focus !disabled", focuscolor),
                    ("readonly", readonly),
                ],
                "darkcolor": [
                    ("focus invalid", self.colors.danger),
                    ("focus !disabled", focuscolor),
                    ("readonly", readonly),
                ],
            },
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_menubutton_style(self, settings, colorname=None):
        """Create a solid style for the ttk.Menubutton widget."""
        STYLE = "TMenubutton"

        foreground = self.colors.get_foreground(colorname)

        if any([colorname is None, not colorname]):
            ttkstyle = STYLE
            background = self.colors.primary
        else:
            ttkstyle = f"{colorname}.{STYLE}"
            background = self.colors.get_color(colorname)

        disabled_bg = self.colorutil.make_transparent(
            0.10, self.colors.fg, self.colors.bg
        )
        disabled_fg = self.colorutil.make_transparent(
            0.30, self.colors.fg, self.colors.bg
        )
        pressed = self.colorutil.make_transparent(
            0.80, background, self.colors.bg
        )
        hover = self.colorutil.make_transparent(
            0.90, background, self.colors.bg
        )

        settings[ttkstyle] = {
            "configure": {
                "foreground": foreground,
                "background": background,
                "bordercolor": background,
                "darkcolor": background,
                "lightcolor": background,
                "arrowsize": self.scale_size(4),
                "arrowcolor": foreground,
                "arrowpadding": (0, 0, 15, 0),
                "relief": tk.RAISED,
                "focusthickness": 0,
                "focuscolor": self.colors.selectfg,
                "padding": (10, 5),
            },
            "map": {
                "arrowcolor": [("disabled", disabled_fg)],
                "foreground": [("disabled", disabled_fg)],
                "background": [
                    ("disabled", disabled_bg),
                    ("pressed !disabled", pressed),
                    ("hover !disabled", hover),
                ],
                "bordercolor": [
                    ("disabled", disabled_bg),
                    ("pressed !disabled", pressed),
                    ("hover !disabled", hover),
                ],
                "darkcolor": [
                    ("disabled", disabled_bg),
                    ("pressed !disabled", pressed),
                    ("hover !disabled", hover),
                ],
                "lightcolor": [
                    ("disabled", disabled_bg),
                    ("pressed !disabled", pressed),
                    ("hover !disabled", hover),
                ],
            },
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_combobox_style(self, settings, colorname=None):
        """Create a style for the ttk.Combobox widget."""
        STYLE = "TCombobox"

        if self.is_light_theme:
            disabled_fg = self.colors.border
            bordercolor = self.colors.border
            readonly = self.colors.light
        else:
            disabled_fg = self.colors.selectbg
            bordercolor = self.colors.selectbg
            readonly = bordercolor

        if any([colorname is None, not colorname]):
            ttkstyle = STYLE
            element = f"{ttkstyle.replace('TC','C')}"
            focuscolor = self.colors.primary
        else:
            ttkstyle = f"{colorname}.{STYLE}"
            element = f"{ttkstyle.replace('TC','C')}"
            focuscolor = self.colors.get_color(colorname)

        if all([colorname, colorname is not None]):
            bordercolor = focuscolor

        ttk_elements = (
            (f"{element}.downarrow", TTK_DEFAULT),
            (f"{element}.padding", TTK_CLAM),
            (f"{element}.textarea", TTK_CLAM),
        )
        for element_name, from_ in ttk_elements:
            settings[element_name] = {
                "element create": ("from", from_),
            }

        settings[ttkstyle] = {
            "configure": {
                "bordercolor": bordercolor,
                "darkcolor": self.colors.inputbg,
                "lightcolor": self.colors.inputbg,
                "arrowcolor": self.colors.inputfg,
                "foreground": self.colors.inputfg,
                "fieldbackground": self.colors.inputbg,
                "background": self.colors.inputbg,
                "insertcolor": self.colors.inputfg,
                "relief": tk.FLAT,
                "padding": 5,
                "arrowsize": self.scale_size(12),
            },
            "map": {
                "background": [("readonly", readonly)],
                "fieldbackground": [("readonly", readonly)],
                "foreground": [("disabled", disabled_fg)],
                "bordercolor": [
                    ("invalid", self.colors.danger),
                    ("focus !disabled", focuscolor),
                    ("hover !disabled", focuscolor),
                ],
                "lightcolor": [
                    ("focus invalid", self.colors.danger),
                    ("focus !disabled", focuscolor),
                    ("pressed !disabled", focuscolor),
                    ("readonly", readonly),
                ],
                "darkcolor": [
                    ("focus invalid", self.colors.danger),
                    ("focus !disabled", focuscolor),
                    ("pressed !disabled", focuscolor),
                    ("readonly", readonly),
                ],
                "arrowcolor": [
                    ("disabled", disabled_fg),
                    ("pressed !disabled", focuscolor),
                    ("focus !disabled", focuscolor),
                    ("hover !disabled", focuscolor),
                ],
            },
            "layout": [
                (
                    "combo.Spinbox.field",
                    {
                        "side": tk.TOP,
                        "sticky": tk.EW,
                        "children": [
                            (
                                "Combobox.downarrow",
                                {"side": tk.RIGHT, "sticky": tk.NS},
                            ),
                            (
                                "Combobox.padding",
                                {
                                    "expand": "1",
                                    "sticky": tk.NSEW,
                                    "children": [
                                        (
                                            "Combobox.textarea",
                                            {"sticky": tk.NSEW},
                                        )
                                    ],
                                },
                            ),
                        ],
                    },
                )
            ],
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_panedwindow_style(self, settings, colorname=None):
        """Create a standard style for the ttk.Panedwindow widget."""
        H_STYLE = "Horizontal.TPanedwindow"
        V_STYLE = "Vertical.TPanedwindow"

        if self.is_light_theme:
            default_color = self.colors.border
        else:
            default_color = self.colors.selectbg

        if any([colorname is None, not colorname]):
            sashcolor = default_color
            h_ttkstyle = H_STYLE
            v_ttkstyle = V_STYLE
        else:
            sashcolor = self.colors.get_color(colorname)
            h_ttkstyle = f"{colorname}.{H_STYLE}"
            v_ttkstyle = f"{colorname}.{V_STYLE}"

        settings["Sash"] = {
            "configure": {"gripcount": 0, "sashthickness": self.scale_size(2)}
        }
        settings[h_ttkstyle] = {
            "configure": {
                "background": sashcolor,
            }
        }
        settings[v_ttkstyle] = {
            "configure": {
                "background": sashcolor,
            }
        }
        if colorname:
            self._register_style(h_ttkstyle)
            self._register_style(v_ttkstyle)

    def create_notebook_style(self, settings, colorname=None):
        """Create a standard style for the ttk.Notebook widget."""
        STYLE = "TNotebook"

        if self.is_light_theme:
            bordercolor = self.colors.border
            foreground = self.colors.inputfg
        else:
            bordercolor = self.colors.selectbg
            foreground = self.colors.selectfg

        if any([colorname is None, not colorname]):
            background = self.colors.inputbg
            selectfg = self.colors.fg
            ttkstyle = STYLE
        else:
            selectfg = self.colors.get_foreground(colorname)
            background = self.colors.get_color(colorname)
            ttkstyle = f"{colorname}.{STYLE}"

        ttkstyle_tab = f"{ttkstyle}.Tab"

        # create widget style
        settings[ttkstyle] = {
            "configure": {
                "background": self.colors.bg,
                "bordercolor": bordercolor,
                "lightcolor": self.colors.bg,
                "darkcolor": self.colors.bg,
                "tabmargins": (0, 1, 1, 0),
            }
        }
        settings[ttkstyle_tab] = {
            "configure": {
                "focuscolor": "",
                "foreground": foreground,
                "padding": (6, 5),
            },
            "map": {
                "background": [
                    ("selected", self.colors.bg),
                    ("!selected", background),
                ],
                "lightcolor": [
                    ("selected", self.colors.bg),
                    ("!selected", background),
                ],
                "bordercolor": [
                    ("selected", bordercolor),
                    ("!selected", bordercolor),
                ],
                "padding": [("selected", (6, 5)), ("!selected", (6, 5))],
                "foreground": [
                    ("selected", foreground),
                    ("!selected", selectfg),
                ],
            },
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_separator_style(self, settings, colorname=None):
        """Create a style for the ttk.Separator widget."""
        HSTYLE = "Horizontal.TSeparator"
        VSTYLE = "Vertical.TSeparator"

        hsize = [20, 1]
        # vsize = [1, 40]

        # style colors
        if self.is_light_theme:
            default_color = self.colors.border
        else:
            default_color = self.colors.selectbg

        if any([colorname is None, not colorname]):
            background = default_color
            h_ttkstyle = HSTYLE
            v_ttkstyle = VSTYLE
        else:
            background = self.colors.get_color(colorname)
            h_ttkstyle = f"{colorname}.{HSTYLE}"
            v_ttkstyle = f"{colorname}.{VSTYLE}"

        # assets
        h_name, v_name = self.assets.create_separator_assets(hsize, background)

        # horizontal separator
        h_element = h_ttkstyle.replace(".TS", ".S")
        h_element_uid = f"{h_element}.separator"
        settings[h_element_uid] = {
            "element create": ("image", h_name),
        }
        settings[h_ttkstyle] = {"layout": [(h_element_uid, {"sticky": tk.EW})]}

        # vertical separator
        v_element = v_ttkstyle.replace(".TS", ".S")
        v_element_uid = f"{v_element}.separator"
        settings[v_element_uid] = {
            "element create": ("image", v_name),
        }
        settings[v_ttkstyle] = {"layout": [(v_element_uid, {"sticky": tk.NS})]}
        if colorname:
            self._register_style(h_ttkstyle)
            self._register_style(v_ttkstyle)

    def create_progressbar_style(self, settings, colorname=None):
        """Create a solid ttk style for the ttk.Progressbar widget."""
        H_STYLE = "Horizontal.TProgressbar"
        V_STYLE = "Vertical.TProgressbar"

        thickness = self.scale_size(10)

        if self.is_light_theme:
            if colorname == LIGHT:
                troughcolor = self.colors.bg
                bordercolor = self.colors.light
            else:
                troughcolor = self.colors.light
                bordercolor = troughcolor
        else:
            troughcolor = self.colorutil.update_hsv(
                self.colors.selectbg, vd=-0.2
            )
            bordercolor = troughcolor

        if any([colorname is None, not colorname]):
            background = self.colors.primary
            h_ttkstyle = H_STYLE
            v_ttkstyle = V_STYLE
        else:
            background = self.colors.get_color(colorname)
            h_ttkstyle = f"{colorname}.{H_STYLE}"
            v_ttkstyle = f"{colorname}.{V_STYLE}"

        # horizontal progressbar
        h_element = h_ttkstyle.replace(".TP", ".P")
        htrough_element = f"{h_element}.trough"
        hpbar_element = f"{h_element}.pbar"
        if htrough_element not in self.existing_elements:
            settings[htrough_element] = {
                "element create": ("from", TTK_CLAM),
            }
            settings[hpbar_element] = {
                "element create": ("from", TTK_DEFAULT),
            }
            self.existing_elements[htrough_element] = True

        settings[h_ttkstyle] = {
            "configure": {
                "thickness": thickness,
                "borderwidth": 1,
                "bordercolor": bordercolor,
                "lightcolor": self.colors.border,
                "pbarrelief": tk.FLAT,
                "troughcolor": troughcolor,
                "background": background,
            },
            "layout": [
                (
                    htrough_element,
                    {
                        "sticky": "nswe",
                        "children": [
                            (hpbar_element, {"side": "left", "sticky": "ns"})
                        ],
                    },
                )
            ],
        }

        # vertical progressbar
        v_element = v_ttkstyle.replace(".TP", ".P")
        vtrough_element = f"{v_element}.trough"
        vpbar_element = f"{v_element}.pbar"
        if vtrough_element not in self.existing_elements:
            settings[vtrough_element] = {
                "element create": ("from", TTK_CLAM),
            }
            settings[vpbar_element] = {
                "element create": ("from", TTK_DEFAULT),
            }
            self.existing_elements[vtrough_element] = True

        settings[v_ttkstyle] = {
            "configure": {
                "thickness": thickness,
                "borderwidth": 1,
                "bordercolor": bordercolor,
                "lightcolor": self.colors.border,
                "pbarrelief": tk.FLAT,
                "troughcolor": troughcolor,
                "background": background,
            },
            "layout": [
                (
                    vtrough_element,
                    {
                        "sticky": "nswe",
                        "children": [
                            (vpbar_element, {"side": "bottom", "sticky": "we"})
                        ],
                    },
                )
            ],
        }
        if colorname:
            self._register_style(h_ttkstyle)
            self._register_style(v_ttkstyle)

    def create_scale_style(self, settings, colorname=None):
        """Create a style for the ttk.Scale widget."""
        STYLE = "TScale"

        if any([colorname is None, not colorname]):
            h_ttkstyle = f"Horizontal.{STYLE}"
            v_ttkstyle = f"Vertical.{STYLE}"
        else:
            h_ttkstyle = f"{colorname}.Horizontal.{STYLE}"
            v_ttkstyle = f"{colorname}.Vertical.{STYLE}"

        # ( normal, pressed, hover, disabled, htrack, vtrack )
        images = self.assets.create_scale_assets(colorname)

        # horizontal scale
        h_element = h_ttkstyle.replace(".TS", ".S")

        settings[f"{h_element}.slider"] = {
            "element create": (
                "image",
                images[0],
                ("disabled", images[3]),
                ("pressed", images[1]),
                ("hover", images[2]),
            ),
        }
        settings[f"{h_element}.track"] = {
            "element create": (
                "image",
                images[4],
            )
        }
        settings[h_ttkstyle] = {
            "layout": [
                (
                    f"{h_element}.focus",
                    {
                        "expand": "1",
                        "sticky": tk.NSEW,
                        "children": [
                            (f"{h_element}.track", {"sticky": tk.EW}),
                            (
                                f"{h_element}.slider",
                                {"side": tk.LEFT, "sticky": ""},
                            ),
                        ],
                    },
                )
            ],
        }
        # vertical scale
        v_element = v_ttkstyle.replace(".TS", ".S")
        settings[f"{v_element}.slider"] = {
            "element create": (
                "image",
                images[0],
                ("disabled", images[3]),
                ("pressed", images[1]),
                ("hover", images[2]),
            ),
        }
        settings[f"{v_element}.track"] = {
            "element create": (
                "image",
                images[5],
            )
        }
        settings[v_ttkstyle] = {
            "layout": [
                (
                    f"{v_element}.focus",
                    {
                        "expand": "1",
                        "sticky": tk.NSEW,
                        "children": [
                            (f"{v_element}.track", {"sticky": tk.NS}),
                            (
                                f"{v_element}.slider",
                                {"side": tk.TOP, "sticky": ""},
                            ),
                        ],
                    },
                )
            ],
        }
        if colorname:
            self._register_style(h_ttkstyle)
            self._register_style(v_ttkstyle)

    def create_scrollbar_style(self, settings, colorname=None):
        """Create a standard style for the ttk.Scrollbar widget."""
        STYLE = "TScrollbar"

        if any([colorname is None, not colorname]):
            h_ttkstyle = f"Horizontal.{STYLE}"
            v_ttkstyle = f"Vertical.{STYLE}"

            if self.is_light_theme:
                background = self.colors.border
            else:
                background = self.colors.selectbg

        else:
            h_ttkstyle = f"{colorname}.Horizontal.{STYLE}"
            v_ttkstyle = f"{colorname}.Vertical.{STYLE}"
            background = self.colors.get_color(colorname)

        if self.is_light_theme:
            if colorname == LIGHT:
                troughcolor = self.colors.bg
            else:
                troughcolor = self.colors.light
        else:
            troughcolor = self.colorutil.update_hsv(
                self.colors.selectbg, vd=-0.2
            )

        pressed = self.colorutil.update_hsv(background, vd=-0.05)
        active = self.colorutil.update_hsv(background, vd=0.05)

        scroll_images = self.assets.create_scrollbar_assets(
            background, pressed, active
        )

        # horizontal scrollbar
        settings[f"{h_ttkstyle}.thumb"] = {
            "element create": (
                "image",
                scroll_images[0],
                ("pressed", scroll_images[1]),
                ("active", scroll_images[2]),
                {"border": (3, 0), "sticky": tk.NSEW},
            )
        }
        settings[h_ttkstyle] = {
            "configure": {
                "troughcolor": troughcolor,
                "darkcolor": troughcolor,
                "bordercolor": troughcolor,
                "lightcolor": troughcolor,
                "arrowcolor": background,
                "arrowsize": self.scale_size(11),
                "background": troughcolor,
                "relief": tk.FLAT,
                "borderwidth": 0,
                "arrowcolor": background,
            },
            "map": {
                "arrowcolor": [("pressed", pressed), ("active", active)],
            },
            "layout": [
                (
                    "Horizontal.Scrollbar.trough",
                    {
                        "sticky": "we",
                        "children": [
                            (
                                "Horizontal.Scrollbar.leftarrow",
                                {"side": "left", "sticky": ""},
                            ),
                            (
                                "Horizontal.Scrollbar.rightarrow",
                                {"side": "right", "sticky": ""},
                            ),
                            (
                                f"{h_ttkstyle}.thumb",
                                {"expand": "1", "sticky": "nswe"},
                            ),
                        ],
                    },
                )
            ],
        }

        # vertical scrollbar
        settings[f"{v_ttkstyle}.thumb"] = {
            "element create": (
                "image",
                scroll_images[3],
                ("pressed", scroll_images[4]),
                ("active", scroll_images[5]),
                {"border": (0, 3), "sticky": tk.NSEW},
            )
        }
        settings[v_ttkstyle] = {
            "configure": {
                "troughcolor": troughcolor,
                "darkcolor": troughcolor,
                "bordercolor": troughcolor,
                "lightcolor": troughcolor,
                "arrowcolor": background,
                "arrowsize": self.scale_size(11),
                "background": troughcolor,
                "relief": tk.FLAT,
                "borderwidth": 0,
                "arrowcolor": background,
            },
            "map": {
                "arrowcolor": [("pressed", pressed), ("active", active)],
            },
            "layout": [
                (
                    "Vertical.Scrollbar.trough",
                    {
                        "sticky": "ns",
                        "children": [
                            (
                                "Vertical.Scrollbar.uparrow",
                                {"side": "top", "sticky": ""},
                            ),
                            (
                                "Vertical.Scrollbar.downarrow",
                                {"side": "bottom", "sticky": ""},
                            ),
                            (
                                f"{v_ttkstyle}.thumb",
                                {"expand": "1", "sticky": "nswe"},
                            ),
                        ],
                    },
                )
            ],
        }
        if colorname:
            self._register_style(h_ttkstyle)
            self._register_style(v_ttkstyle)

    def create_spinbox_style(self, settings, colorname=None):
        """Create a style for the ttk.Spinbox widget."""
        STYLE = "TSpinbox"

        if self.is_light_theme:
            disabled_fg = self.colors.border
            bordercolor = self.colors.border
            readonly = self.colors.light
        else:
            disabled_fg = self.colors.selectbg
            bordercolor = self.colors.selectbg
            readonly = bordercolor

        if any([colorname is None, not colorname]):
            ttkstyle = STYLE
            focuscolor = self.colors.primary
        else:
            ttkstyle = f"{colorname}.{STYLE}"
            focuscolor = self.colors.get_color(colorname)

        if all([colorname, colorname is not None]):
            bordercolor = focuscolor

        if colorname == "light":
            arrowfocus = self.colors.fg
        else:
            arrowfocus = focuscolor

        element = ttkstyle.replace(".TS", ".S")
        settings[f"{element}.uparrow"] = {
            "element create": ("from", TTK_DEFAULT),
        }
        settings[f"{element}.downarrow"] = {
            "element create": ("from", TTK_DEFAULT),
        }
        settings[ttkstyle] = {
            "configure": {
                "bordercolor": bordercolor,
                "darkcolor": self.colors.inputbg,
                "lightcolor": self.colors.inputbg,
                "fieldbackground": self.colors.inputbg,
                "foreground": self.colors.inputfg,
                "borderwidth": 0,
                "background": self.colors.inputbg,
                "relief": tk.FLAT,
                "arrowcolor": self.colors.inputfg,
                "insertcolor": self.colors.inputfg,
                "arrowsize": self.scale_size(12),
                "padding": (10, 5),
            },
            "map": {
                "foreground": [("disabled", disabled_fg)],
                "fieldbackground": [("readonly", readonly)],
                "background": [("readonly", readonly)],
                "lightcolor": [
                    ("focus invalid", self.colors.danger),
                    ("focus !disabled", focuscolor),
                    ("readonly", readonly),
                ],
                "darkcolor": [
                    ("focus invalid", self.colors.danger),
                    ("focus !disabled", focuscolor),
                    ("readonly", readonly),
                ],
                "bordercolor": [
                    ("invalid", self.colors.danger),
                    ("focus !disabled", focuscolor),
                    ("hover !disabled", focuscolor),
                ],
                "arrowcolor": [
                    ("disabled !disabled", disabled_fg),
                    ("pressed !disabled", arrowfocus),
                    ("hover !disabled", arrowfocus),
                ],
            },
            "layout": [
                (
                    f"{element}.field",
                    {
                        "side": tk.TOP,
                        "sticky": tk.EW,
                        "children": [
                            (
                                "null",
                                {
                                    "side": tk.RIGHT,
                                    "sticky": "",
                                    "children": [
                                        (
                                            f"{element}.uparrow",
                                            {"side": tk.TOP, "sticky": tk.E},
                                        ),
                                        (
                                            f"{element}.downarrow",
                                            {
                                                "side": tk.BOTTOM,
                                                "sticky": tk.E,
                                            },
                                        ),
                                    ],
                                },
                            ),
                            (
                                f"{element}.padding",
                                {
                                    "sticky": tk.NSEW,
                                    "children": [
                                        (
                                            f"{element}.textarea",
                                            {"sticky": tk.NSEW},
                                        )
                                    ],
                                },
                            ),
                        ],
                    },
                )
            ],
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_treeview_style(self, settings, colorname=None):
        """Create a style for the ttk.Treeview widget."""
        STYLE = "Treeview"

        f = font.nametofont("TkDefaultFont")
        rowheight = f.metrics()["linespace"]

        if self.is_light_theme:
            disabled_fg = self.colorutil.update_hsv(
                self.colors.inputbg, vd=-0.2
            )
            bordercolor = self.colors.border
        else:
            disabled_fg = self.colorutil.update_hsv(
                self.colors.inputbg, vd=-0.3
            )
            bordercolor = self.colors.selectbg

        if any([colorname is None, not colorname]):
            background = self.colors.inputbg
            foreground = self.colors.inputfg
            body_style = STYLE
            header_style = f"{STYLE}.Heading"
            focuscolor = self.colors.primary
        elif colorname == LIGHT and self.is_light_theme:
            background = self.colors.get_color(colorname)
            foreground = self.colors.fg
            body_style = f"{colorname}.{STYLE}"
            header_style = f"{colorname}.{STYLE}.Heading"
            focuscolor = background
            bordercolor = focuscolor
        else:
            background = self.colors.get_color(colorname)
            foreground = self.colors.selectfg
            body_style = f"{colorname}.{STYLE}"
            header_style = f"{colorname}.{STYLE}.Heading"
            focuscolor = background
            bordercolor = focuscolor

        # treeview header
        settings[header_style] = {
            "configure": {
                "background": background,
                "foreground": foreground,
                "relief": tk.FLAT,
                "padding": 5,
            },
            "map": {
                "foreground": [("disabled", disabled_fg)],
                "bordercolor": [("focus !disabled", background)],
            },
        }
        # treeview body
        settings[body_style] = {
            "configure": {
                "background": self.colors.inputbg,
                "fieldbackground": self.colors.inputbg,
                "foreground": self.colors.inputfg,
                "bordercolor": bordercolor,
                "lightcolor": self.colors.inputbg,
                "darkcolor": self.colors.inputbg,
                "borderwidth": 2,
                "padding": 0,
                "rowheight": rowheight,
                "relief": tk.RAISED,
            },
            "map": {
                "background": [("selected", self.colors.selectbg)],
                "foreground": [
                    ("disabled", disabled_fg),
                    ("selected", self.colors.selectfg),
                ],
                "bordercolor": [
                    ("disabled", bordercolor),
                    ("focus", focuscolor),
                    ("pressed", focuscolor),
                    ("hover", focuscolor),
                ],
                "lightcolor": [("focus", focuscolor)],
                "darkcolor": [("focus", focuscolor)],
            },
            "layout": [
                (
                    "Button.border",
                    {
                        "sticky": tk.NSEW,
                        "border": "1",
                        "children": [
                            (
                                "Treeview.padding",
                                {
                                    "sticky": tk.NSEW,
                                    "children": [
                                        (
                                            "Treeview.treearea",
                                            {"sticky": tk.NSEW},
                                        )
                                    ],
                                },
                            )
                        ],
                    },
                )
            ],
        }
        # try:
        # self.style.element_create("Treeitem.indicator", "from", TTK_ALT)
        # except:
        # pass
        settings["Treeitem.indicator"] = {
            "element create": ("from", TTK_ALT),
        }
        if colorname:
            self._register_style(body_style)

    def create_radiobutton_style(self, settings, colorname=None):
        """Create a style for the ttk.Radiobutton widget."""

        STYLE = "TRadiobutton"

        disabled_fg = self.colorutil.make_transparent(
            0.30, self.colors.fg, self.colors.bg
        )

        if any([colorname is None, not colorname]):
            ttkstyle = STYLE
            colorname = PRIMARY
        else:
            ttkstyle = f"{colorname}.{STYLE}"

        # ( off, on, disabled )
        images = self.assets.create_radiobutton_assets(colorname)
        width = self.scale_size(20)
        borderpad = self.scale_size(4)

        settings[f"{ttkstyle}.indicator"] = {
            "element create": (
                "image",
                images[1],
                ("disabled selected", images[3]),
                ("disabled", images[2]),
                ("!selected", images[0]),
                {
                    "width": width,
                    "border": borderpad,
                    "sticky": tk.W,
                },
            )
        }

        settings[ttkstyle] = {
            "map": {"foreground": [("disabled", disabled_fg)]},
            "layout": [
                (
                    "Radiobutton.padding",
                    {
                        "children": [
                            (
                                f"{ttkstyle}.indicator",
                                {"side": tk.LEFT, "sticky": ""},
                            ),
                            (
                                "Radiobutton.focus",
                                {
                                    "children": [
                                        (
                                            "Radiobutton.label",
                                            {"sticky": tk.NSEW},
                                        )
                                    ],
                                    "side": tk.LEFT,
                                    "sticky": "",
                                },
                            ),
                        ],
                        "sticky": tk.NSEW,
                    },
                )
            ],
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_checkbutton_style(self, settings, colorname=None):
        """Create a standard style for the ttk.Checkbutton widget."""
        STYLE = "TCheckbutton"

        disabled_fg = self.colorutil.make_transparent(
            0.3, self.colors.fg, self.colors.bg
        )

        if any([colorname is None, not colorname]):
            colorname = PRIMARY
            ttkstyle = STYLE
        else:
            ttkstyle = f"{colorname}.TCheckbutton"

        # ( off, on, disabled )
        images = self.assets.create_checkbutton_assets(colorname)

        element = ttkstyle.replace(".TC", ".C")
        width = self.scale_size(20)
        borderpad = self.scale_size(4)

        settings[f"{element}.indicator"] = {
            "element create": (
                "image",
                images[1],
                ("disabled selected", images[4]),
                ("disabled alternate", images[5]),
                ("disabled", images[2]),
                ("alternate", images[3]),
                ("!selected", images[0]),
                {
                    "width": width,
                    "border": borderpad,
                    "sticky": tk.W,
                },
            )
        }
        settings[ttkstyle] = {
            "configure": {"foreground": self.colors.fg},
            "map": {"foreground": [("disabled", disabled_fg)]},
            "layout": [
                (
                    "Checkbutton.padding",
                    {
                        "children": [
                            (
                                f"{element}.indicator",
                                {"side": tk.LEFT, "sticky": ""},
                            ),
                            (
                                "Checkbutton.focus",
                                {
                                    "children": [
                                        (
                                            "Checkbutton.label",
                                            {"sticky": tk.NSEW},
                                        )
                                    ],
                                    "side": tk.LEFT,
                                    "sticky": "",
                                },
                            ),
                        ],
                        "sticky": tk.NSEW,
                    },
                )
            ],
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_sizegrip_style(self, settings, colorname=None):
        """Create a style for the ttk.Sizegrip widget."""
        STYLE = "TSizegrip"

        if any([colorname is None, not colorname]):
            ttkstyle = STYLE

            if self.is_light_theme:
                grip_color = self.colors.border
            else:
                grip_color = self.colors.inputbg
        else:
            ttkstyle = f"{colorname}.{STYLE}"
            grip_color = self.colors.get_color(colorname)

        image = self.assets.create_sizegrip_assets(grip_color)

        settings[f"{ttkstyle}.Sizegrip.sizegrip"] = {
            "element create": ("image", image)
        }
        settings[ttkstyle] = {
            "layout": [
                (
                    f"{ttkstyle}.Sizegrip.sizegrip",
                    {"side": tk.BOTTOM, "sticky": tk.SE},
                )
            ],
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_striped_progressbar_style(self, settings, colorname=None):
        """Create a striped style for the ttk.Progressbar widget.

        Parameters:

            colorname (str):
                The primary widget color label.
        """
        HSTYLE = "Striped.Horizontal.TProgressbar"
        VSTYLE = "Striped.Vertical.TProgressbar"

        thickness = self.scale_size(12)

        if any([colorname is None, colorname == ""]):
            h_ttkstyle = HSTYLE
            v_ttkstyle = VSTYLE
        else:
            h_ttkstyle = f"{colorname}.{HSTYLE}"
            v_ttkstyle = f"{colorname}.{VSTYLE}"

        if self.is_light_theme:
            if colorname == LIGHT:
                troughcolor = self.colors.bg
                bordercolor = self.colors.light
            else:
                troughcolor = self.colors.light
                bordercolor = troughcolor
        else:
            troughcolor = self.colorutil.update_hsv(
                self.colors.selectbg, vd=-0.2
            )
            bordercolor = troughcolor

        # ( horizontal, vertical )
        images = self.assets.create_striped_progressbar_assets(
            thickness, colorname
        )

        # horizontal progressbar
        h_element = h_ttkstyle.replace(".TP", ".P")
        settings[f"{h_element}.pbar"] = {
            "element create": (
                "image",
                images[0],
                {
                    "width": thickness,
                    "sticky": tk.EW,
                },
            )
        }
        settings[h_ttkstyle] = {
            "configure": {
                "troughcolor": troughcolor,
                "thickness": thickness,
                "bordercolor": bordercolor,
                "borderwidth": 1,
            },
            "layout": [
                (
                    f"{h_element}.trough",
                    {
                        "sticky": tk.NSEW,
                        "children": [
                            (
                                f"{h_element}.pbar",
                                {"side": tk.LEFT, "sticky": tk.NS},
                            )
                        ],
                    },
                )
            ],
        }

        # vertical progressbar
        v_element = v_ttkstyle.replace(".TP", ".P")
        settings[f"{v_element}.pbar"] = {
            "element create": (
                "image",
                images[1],
                {
                    "width": thickness,
                    "sticky": tk.NS,
                },
            )
        }
        settings[v_ttkstyle] = {
            "configure": {
                "troughcolor": troughcolor,
                "thickness": thickness,
                "bordercolor": bordercolor,
                "borderwidth": 1,
            },
            "layout": [
                (
                    f"{v_element}.trough",
                    {
                        "sticky": tk.NSEW,
                        "children": [
                            (
                                f"{v_element}.pbar",
                                {"side": tk.BOTTOM, "sticky": tk.EW},
                            )
                        ],
                    },
                )
            ],
        }
        if colorname:
            self._register_style(h_ttkstyle)
            self._register_style(v_ttkstyle)

    def create_square_toggle_style(self, settings, colorname=None):
        """Create a square toggle style for the ttk.Checkbutton widget.

        Parameters:

            colorname (str):
                The color label used to style the widget.
        """

        STYLE = "Square.Toggle"

        disabled_fg = self.colorutil.make_transparent(
            0.30, self.colors.fg, self.colors.bg
        )

        if any([colorname is None, colorname == ""]):
            ttkstyle = STYLE
        else:
            ttkstyle = f"{colorname}.{STYLE}"

        # ( off, on, disabled )
        images = self.assets.create_square_toggle_assets(colorname)

        width = self.scale_size(28)
        borderpad = self.scale_size(4)

        settings[f"{ttkstyle}.indicator"] = {
            "element create": (
                "image",
                images[1],
                ("disabled selected", images[3]),
                ("disabled", images[2]),
                ("!selected", images[0]),
                {
                    "width": width,
                    "sticky": tk.W,
                    "border": borderpad,
                },
            )
        }
        settings[ttkstyle] = {
            "configure": {
                "relief": tk.FLAT,
                "borderwidth": 0,
                "foreground": self.colors.fg,
            },
            "layout": [
                (
                    "Toolbutton.border",
                    {
                        "sticky": tk.NSEW,
                        "children": [
                            (
                                "Toolbutton.padding",
                                {
                                    "sticky": tk.NSEW,
                                    "children": [
                                        (
                                            f"{ttkstyle}.indicator",
                                            {"side": tk.LEFT},
                                        ),
                                        (
                                            "Toolbutton.label",
                                            {"side": tk.LEFT},
                                        ),
                                    ],
                                },
                            )
                        ],
                    },
                )
            ],
            "map": {
                "foreground": [("disabled", disabled_fg)],
                "background": [
                    ("selected", self.colors.bg),
                    ("!selected", self.colors.bg),
                ],
            },
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_round_toggle_style(self, settings, colorname=None):
        """Create a round toggle style for the ttk.Checkbutton widget.

        Parameters:

            colorname (str):
                The color label used to style the widget.
        """
        STYLE = "Round.Toggle"

        disabled_fg = self.colorutil.make_transparent(
            0.30, self.colors.fg, self.colors.bg
        )

        if any([colorname is None, colorname == ""]):
            ttkstyle = STYLE
            colorname = PRIMARY
        else:
            ttkstyle = f"{colorname}.{STYLE}"

        # ( off, on, disabled )
        images = self.assets.create_round_toggle_assets(colorname)

        indicator_element = f"{ttkstyle}.indicator"
        if indicator_element not in self.existing_elements:
            width = self.scale_size(28)
            borderpad = self.scale_size(4)

            settings[indicator_element] = {
                "element create": (
                    "image",
                    images[1],
                    ("disabled selected", images[3]),
                    ("disabled", images[2]),
                    ("!selected", images[0]),
                    {
                        "width": width,
                        "sticky": tk.W,
                        "border": borderpad,
                    },
                )
            }
            self.existing_elements[indicator_element] = True

        settings[ttkstyle] = {
            "configure": {
                "relief": tk.FLAT,
                "borderwidth": 0,
                "padding": 0,
                "foreground": self.colors.fg,
                "background": self.colors.bg,
            },
            "map": {
                "foreground": [("disabled", disabled_fg)],
                "background": [("selected", self.colors.bg)],
            },
            "layout": [
                (
                    "Toolbutton.border",
                    {
                        "sticky": tk.NSEW,
                        "children": [
                            (
                                "Toolbutton.padding",
                                {
                                    "sticky": tk.NSEW,
                                    "children": [
                                        (
                                            f"{ttkstyle}.indicator",
                                            {"side": tk.LEFT},
                                        ),
                                        (
                                            "Toolbutton.label",
                                            {"side": tk.LEFT},
                                        ),
                                    ],
                                },
                            )
                        ],
                    },
                )
            ],
        }
        if colorname:
            self._register_style(ttkstyle)

    def create_round_scrollbar_style(self, settings, colorname=None):
        """Create a round style for the ttk.Scrollbar widget.

        Parameters:

            colorname (str):
                The color label used to style the widget.
        """
        STYLE = "TScrollbar"

        if any([colorname is None, colorname == ""]):
            h_ttkstyle = f"Round.Horizontal.{STYLE}"
            v_ttkstyle = f"Round.Vertical.{STYLE}"

            if self.is_light_theme:
                background = self.colors.border
            else:
                background = self.colors.selectbg

        else:
            h_ttkstyle = f"{colorname}.Round.Horizontal.{STYLE}"
            v_ttkstyle = f"{colorname}.Round.Vertical.{STYLE}"
            background = self.colors.get_color(colorname)

        if self.is_light_theme:
            if colorname == LIGHT:
                troughcolor = self.colors.bg
            else:
                troughcolor = self.colors.light
        else:
            troughcolor = self.colorutil.update_hsv(
                self.colors.selectbg, vd=-0.2
            )

        pressed = self.colorutil.update_hsv(background, vd=-0.05)
        active = self.colorutil.update_hsv(background, vd=0.05)

        scroll_images = self.assets.create_round_scrollbar_assets(
            background, pressed, active
        )

        # horizontal scrollbar
        settings[f"{h_ttkstyle}.thumb"] = {
            "element create": (
                "image",
                scroll_images[0],
                ("pressed", scroll_images[1]),
                ("active", scroll_images[2]),
                {
                    "padding": 0,
                    "sticky": tk.EW,
                    "border": self.scale_size(9),
                },
            )
        }
        settings[h_ttkstyle] = {
            "configure": {
                "troughcolor": troughcolor,
                "darkcolor": troughcolor,
                "bordercolor": troughcolor,
                "lightcolor": troughcolor,
                "arrowcolor": background,
                "arrowsize": self.scale_size(11),
                "background": troughcolor,
                "relief": tk.FLAT,
                "borderwidth": 0,
            },
            "layout": [
                (
                    "Horizontal.Scrollbar.trough",
                    {
                        "sticky": "we",
                        "children": [
                            (
                                "Horizontal.Scrollbar.leftarrow",
                                {"side": "left", "sticky": ""},
                            ),
                            (
                                "Horizontal.Scrollbar.rightarrow",
                                {"side": "right", "sticky": ""},
                            ),
                            (
                                f"{h_ttkstyle}.thumb",
                                {"expand": "1", "sticky": "nswe"},
                            ),
                        ],
                    },
                )
            ],
            "map": {"arrowcolor": [("pressed", pressed), ("active", active)]},
        }

        # vertical scrollbar
        settings[f"{v_ttkstyle}.thumb"] = {
            "element create": (
                "image",
                scroll_images[3],
                ("pressed", scroll_images[4]),
                ("active", scroll_images[5]),
                {
                    "padding": 0,
                    "sticky": tk.NS,
                    "border": self.scale_size(9),
                },
            )
        }
        settings[v_ttkstyle] = {
            "configure": {
                "troughcolor": troughcolor,
                "darkcolor": troughcolor,
                "bordercolor": troughcolor,
                "lightcolor": troughcolor,
                "arrowcolor": background,
                "arrowsize": self.scale_size(11),
                "background": troughcolor,
                "relief": tk.FLAT,
                # "borderwidth": 0,
            },
            "layout": [
                (
                    "Vertical.Scrollbar.trough",
                    {
                        "sticky": "ns",
                        "children": [
                            (
                                "Vertical.Scrollbar.uparrow",
                                {"side": "top", "sticky": ""},
                            ),
                            (
                                "Vertical.Scrollbar.downarrow",
                                {"side": "bottom", "sticky": ""},
                            ),
                            (
                                f"{v_ttkstyle}.thumb",
                                {"expand": "1", "sticky": "nswe"},
                            ),
                        ],
                    },
                )
            ],
            "map": {"arrowcolor": [("pressed", pressed), ("active", active)]},
        }
        if colorname:
            self._register_style(h_ttkstyle)
            self._register_style(v_ttkstyle)

    def create_table_treeview_style(self, settings, colorname=None):
        """Create a style for the Tableview widget.

        Parameters:

            colorname (str):
                The color label used to style the widget.
        """
        STYLE = "Table.Treeview"

        f = font.nametofont("TkDefaultFont")
        rowheight = f.metrics()["linespace"]

        if self.is_light_theme:
            disabled_fg = self.colorutil.update_hsv(
                self.colors.inputbg, vd=-0.2
            )
            bordercolor = self.colors.border
            hover = self.colorutil.update_hsv(self.colors.light, vd=-0.1)
        else:
            disabled_fg = self.colorutil.update_hsv(
                self.colors.inputbg, vd=-0.3
            )
            bordercolor = self.colors.selectbg
            hover = self.colorutil.update_hsv(self.colors.dark, vd=0.1)

        if any([colorname is None, colorname == ""]):
            background = self.colors.inputbg
            foreground = self.colors.inputfg
            body_style = STYLE
            header_style = f"{STYLE}.Heading"
        elif colorname == LIGHT and self.is_light_theme:
            background = self.colors.get_color(colorname)
            foreground = self.colors.fg
            body_style = f"{colorname}.{STYLE}"
            header_style = f"{colorname}.{STYLE}.Heading"
            hover = self.colorutil.update_hsv(background, vd=-0.1)
        else:
            background = self.colors.get_color(colorname)
            foreground = self.colors.selectfg
            body_style = f"{colorname}.{STYLE}"
            header_style = f"{colorname}.{STYLE}.Heading"
            hover = self.colorutil.update_hsv(background, vd=0.1)

        # treeview header
        settings[header_style] = {
            "configure": {
                "background": background,
                "foreground": foreground,
                "relief": tk.RAISED,
                "borderwidth": 1,
                "darkcolor": background,
                "bordercolor": bordercolor,
                "lightcolor": background,
                "padding": 5,
            },
            "map": {
                "foreground": [("disabled", disabled_fg)],
                "background": [
                    ("active !disabled", hover),
                ],
                "darkcolor": [
                    ("active !disabled", hover),
                ],
                "lightcolor": [
                    ("active !disabled", hover),
                ],
            },
        }

        settings[body_style] = {
            "configure": {
                "background": self.colors.inputbg,
                "fieldbackground": self.colors.inputbg,
                "foreground": self.colors.inputfg,
                "bordercolor": bordercolor,
                "lightcolor": self.colors.inputbg,
                "darkcolor": self.colors.inputbg,
                "borderwidth": 2,
                "padding": 0,
                "rowheight": rowheight,
                "relief": tk.RAISED,
            },
            "map": {
                "background": [("selected", self.colors.selectbg)],
                "foreground": [
                    ("disabled", disabled_fg),
                    ("selected", self.colors.selectfg),
                ],
            },
            "layout": [
                (
                    "Button.border",
                    {
                        "sticky": tk.NSEW,
                        "border": "1",
                        "children": [
                            (
                                "Treeview.padding",
                                {
                                    "sticky": tk.NSEW,
                                    "children": [
                                        (
                                            "Treeview.treearea",
                                            {"sticky": tk.NSEW},
                                        )
                                    ],
                                },
                            )
                        ],
                    },
                )
            ],
        }
        if colorname:
            self._register_style(body_style)



# encoding: utf-8
import support
import pygubu
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk

import fixpath


class TestEntry(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = """<?xml version="1.0" ?>
<interface>
  <object class="ttk.Frame" id="mainwindow">
    <property name="height">250</property>
    <property name="width">250</property>
    <layout>
      <property name="column">0</property>
      <property name="sticky">nsew</property>
      <property name="propagate">True</property>
      <property name="row">0</property>
    </layout>
    <child>
      <object class="ttk.Label" id="label">
        <property name="anchor">e</property>
        <property name="background">#94f900</property>
        <property name="borderwidth">2</property>
        <property name="compound">right</property>
        <property name="foreground">#690400</property>
        <property name="padding">2</property>
        <property name="relief">ridge</property>
        <property name="text">-- A Label --</property>
        <property name="textvariable">label_var</property>
        <property name="width">20</property>
        <property name="justify">right</property>
        <layout>
          <property name="column">0</property>
          <property name="propagate">True</property>
          <property name="row">0</property>
        </layout>
      </object>
    </child>
  </object>
</interface>
"""
        self.builder = builder = pygubu.Builder()
        builder.add_from_string(xmldata)
        self.widget = builder.get_object("label")

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.assertIsInstance(self.widget, ttk.Label)
        self.widget.destroy()

    def test_text(self):
        prop = "text"
        expected_value = "-- A Label --"
        tclobj = self.widget.cget(prop)
        value = str(tclobj)
        self.assertEqual(expected_value, value)
        self.widget.destroy()

    def test_textvariable(self):
        varname = "label_var"
        expected_value = "-- A Label --"
        var = self.builder.get_variable(varname)
        self.assertIsInstance(var, tk.StringVar)
        self.assertEqual(expected_value, var.get())

        newlabel = "Changed"
        var.set(newlabel)
        self.assertEqual(newlabel, self.widget.cget("text"))
        self.widget.destroy()

    def test_justify(self):
        prop = "justify"
        expected_value = "right"
        tclobj = self.widget.cget(prop)
        value = str(tclobj)
        self.assertEqual(expected_value, value)
        self.widget.destroy()

    def test_anchor(self):
        prop = "anchor"
        expected_value = "e"
        tclobj = self.widget.cget(prop)
        value = str(tclobj)
        self.assertEqual(expected_value, value)
        self.widget.destroy()

    def test_background(self):
        prop = "background"
        expected_value = "#94f900"
        tclobj = self.widget.cget(prop)
        value = str(tclobj)
        self.assertEqual(expected_value, value)
        self.widget.destroy()

    def test_borderwidth(self):
        prop = "borderwidth"
        expected_value = "2"
        tclobj = self.widget.cget(prop)
        value = str(tclobj)
        self.assertEqual(expected_value, value)
        self.widget.destroy()

    def test_compound(self):
        prop = "compound"
        expected_value = "right"
        tclobj = self.widget.cget(prop)
        value = str(tclobj)
        self.assertEqual(expected_value, value)
        self.widget.destroy()

    def test_foreground(self):
        prop = "foreground"
        expected_value = "#690400"
        tclobj = self.widget.cget(prop)
        value = str(tclobj)
        self.assertEqual(expected_value, value)
        self.widget.destroy()

    def test_padding(self):
        prop = "padding"
        expected_value = "2"
        tclobj = self.widget.cget(prop)[0]
        value = str(tclobj)
        self.assertEqual(expected_value, value)
        self.widget.destroy()

    def test_relief(self):
        prop = "relief"
        expected_value = "ridge"
        tclobj = self.widget.cget(prop)
        value = str(tclobj)
        self.assertEqual(expected_value, value)
        self.widget.destroy()

    def test_width(self):
        prop = "width"
        expected_value = "20"
        tclobj = self.widget.cget(prop)
        value = str(tclobj)
        self.assertEqual(expected_value, value)
        self.widget.destroy()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3
import tkinter as tk
import tkinter.ttk as ttk


#
# Base class definition
#
class ColorInputUI(ttk.Frame):
    def __init__(self, master=None, **kw):
        super().__init__(master, **kw)
        self._frame = tk.Frame(self, name="_frame")
        self._frame.configure(borderwidth=0, height=10, relief="flat", width=24)
        self._frame.pack(fill="y", side="left")
        self._entry = ttk.Entry(self, name="_entry")
        self._entry.configure(validate="all")
        self._entry.pack(expand=True, fill="both", padx=2, side="left")
        self._entry.bind("<FocusOut>", self.on_focusout, add="")
        self._entry.bind("<KeyPress>", self.on_keypress, add="")
        self._button = ttk.Button(self, name="_button")
        self._button.configure(
            compound="center",
            style="Toolbutton",
            takefocus=True,
            text="…",
            width=-2,
        )
        self._button.pack(fill="both", side="left")
        self._button.configure(command=self.on_picker_clicked)
        self.configure(height=25, width=100)
        # self.pack(side="top")

    def on_focusout(self, event=None):
        pass

    def on_keypress(self, event=None):
        pass

    def on_picker_clicked(self):
        pass


if __name__ == "__main__":
    root = tk.Tk()
    widget = ColorInputUI(root)
    widget.pack(expand=True, fill="both")
    root.mainloop()



hiddenimports = [
    'pygubu.plugins.tk.tkstdwidgets',
    'pygubu.plugins.ttk.ttkstdwidgets',
    'pygubu.plugins.pygubu.dialog',
    'pygubu.plugins.pygubu.editabletreeview',
    'pygubu.plugins.pygubu.scrollbarhelper',
    'pygubu.plugins.pygubu.scrolledframe',
    'pygubu.plugins.pygubu.tkscrollbarhelper',
    'pygubu.plugins.pygubu.tkscrolledframe',
    'pygubu.plugins.pygubu.pathchooserinput',
]



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk

import fixpath
import pygubu
import support


class TestFrame(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_frame.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("mainwindow")

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.assertIsInstance(self.widget, ttk.Frame)
        self.widget.destroy()

    def test_padding(self):
        tclobj = self.widget.cget("padding")[0]
        padding = str(tclobj)
        self.assertEqual("10", padding)
        self.widget.destroy()

    def test_width(self):
        tclobj = self.widget.cget("width")
        width = str(tclobj)
        self.assertEqual("250", width)
        self.widget.destroy()

    def test_class_(self):
        tclobj = self.widget.cget("class")
        class_ = str(tclobj)
        self.assertEqual("MyCustomFrame", class_)
        self.widget.destroy()

    def test_relief(self):
        tclobj = self.widget.cget("relief")
        relief = str(tclobj)
        self.assertEqual(tk.SUNKEN, relief)
        self.widget.destroy()

    def test_style(self):
        tclobj = self.widget.cget("style")
        style = str(tclobj)
        self.assertEqual("MyFrameStyle.TFrame", style)
        self.widget.destroy()

    def test_takefocus(self):
        tclobj = self.widget.cget("takefocus")
        takefocus = str(tclobj)
        self.assertEqual("1", takefocus)
        self.widget.destroy()

    def test_cursor(self):
        tclobj = self.widget.cget("cursor")
        cursor = str(tclobj)
        self.assertEqual("cross", cursor)
        self.widget.destroy()

    def test_layout(self):
        ginfo = self.widget.grid_info()
        expected = [
            ("row", "0"),
            ("column", "0"),
            ("sticky", "nesw"),
            ("pady", "10"),
            ("padx", "5"),
            ("ipadx", "2"),
            ("ipady", "4"),
            ("rowspan", "1"),
            ("columnspan", "2"),
        ]
        for k, ev in expected:
            value = str(ginfo[k])
            self.assertEqual(value, ev)

        # FIX TEST: since interface v1.1 propagate is applied to parent? NO!
        propagate = self.widget.grid_propagate()
        self.assertEqual(None, propagate)
        self.widget.destroy()

    def test_child_count(self):
        count = len(self.widget.children)
        self.assertEqual(1, count)
        self.widget.destroy()

    def test_binding_dict(self):
        success = []

        def on_button_click(event):
            success.append(1)

        def on_button_click2(event):
            success.append(1)

        cbdic = {
            "on_button_click": on_button_click,
            "on_button_click2": on_button_click2,
        }
        self.builder.connect_callbacks(cbdic)

        support.simulate_mouse_click(self.widget, 5, 5)
        self.widget.update_idletasks()

        self.assertTrue(success)
        self.widget.destroy()

    def test_binding_object(self):
        success = []

        class AnObject:
            def on_button_click(self, event):
                success.append(1)

            def on_button_click2(self, event):
                success.append(1)

        cbobj = AnObject()
        self.builder.connect_callbacks(cbobj)

        support.simulate_mouse_click(self.widget, 5, 5)
        self.widget.update_idletasks()

        self.assertTrue(success)
        self.widget.destroy()

    def test_binding_add(self):
        success = []

        def on_button_click(event):
            success.append(1)

        def on_button_click2(event):
            success.append(1)

        cbdic = {
            "on_button_click": on_button_click,
            "on_button_click2": on_button_click2,
        }
        self.builder.connect_callbacks(cbdic)

        support.simulate_mouse_click(self.widget, 5, 5)
        self.widget.update_idletasks()

        self.assertTrue(len(success) == 2)
        self.widget.destroy()


if __name__ == "__main__":
    unittest.main()



import awesometkinter as atk
from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.tk.tkstdwidgets import TKLabel
from ..awesometkinter import _designer_tab_label, _plugin_uid


class AutoWrappingLabelBO(TKLabel):
    class_ = atk.label.AutoWrappingLabel


_builder_uid = _plugin_uid + ".AutoWrappingLabel"
register_widget(
    _builder_uid,
    AutoWrappingLabelBO,
    "AutoWrappingLabel",
    ("ttk", _designer_tab_label),
    group=1,
)


class AutofitLabelBO(TKLabel):
    OPTIONS_CUSTOM = ("refresh_time",)
    properties = (
        TKLabel.OPTIONS_STANDARD + TKLabel.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = TKLabel.ro_properties + OPTIONS_CUSTOM
    class_ = atk.label.AutofitLabel


_builder_uid = _plugin_uid + ".AutofitLabel"
register_widget(
    _builder_uid,
    AutofitLabelBO,
    "AutofitLabel",
    ("ttk", _designer_tab_label),
    group=1,
)

register_custom_property(
    _builder_uid, "refresh_time", "naturalnumber", help=_("milliseconds")
)



#!/usr/bin/python3
import tkinter as tk
import tkinter.ttk as ttk
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
)
from pygubu.widgets.colorinput import ColorInput
from pygubu.i18n import _


#
# Builder definition section
#


class ColorInputBO(BuilderObject):
    class_ = ColorInput


_builder_id = "pygubu.widgets.ColorInput"
register_widget(
    _builder_id, ColorInputBO, "ColorInput", ("ttk", _("Pygubu Widgets"))
)



import importlib
from pygubu.i18n import _
from pygubu.api.v1 import BuilderLoaderPlugin

_designer_tab_label = _("AwesomeTkinter")
_plugin_uid = "awesometkinter"


class AwesometkinterLoader(BuilderLoaderPlugin):
    module_map = {
        "pygubu.plugins.awesometkinter.frame": (
            f"{_plugin_uid}.Frame3d",
            f"{_plugin_uid}.ScrollableFrame",
        ),
        "pygubu.plugins.awesometkinter.button": (
            f"{_plugin_uid}.Button3d",
            f"{_plugin_uid}.Radiobutton",
            f"{_plugin_uid}.Checkbutton",
        ),
        "pygubu.plugins.awesometkinter.label": (
            f"{_plugin_uid}.AutoWrappingLabel",
            f"{_plugin_uid}.AutofitLabel",
        ),
        "pygubu.plugins.awesometkinter.progressbar": (
            f"{_plugin_uid}.RadialProgressbar",
            f"{_plugin_uid}.RadialProgressbar3d",
            f"{_plugin_uid}.Segmentbar",
        ),
        "pygubu.plugins.awesometkinter.scrollbar": (
            f"{_plugin_uid}.SimpleScrollbar",
        ),
        "pygubu.plugins.awesometkinter.text": (f"{_plugin_uid}.ScrolledText",),
    }

    def do_activate(self) -> bool:
        spec = importlib.util.find_spec("awesometkinter")
        return True if spec is not None else False

    def get_module_for(self, identifier: str) -> str:
        for module, identifiers in self.module_map.items():
            if identifier in identifiers:
                return module
        return None

    def get_all_modules(self):
        return [m for m in self.module_map.keys()]

    def can_load(self, identifier: str) -> bool:
        return identifier.startswith("awesometkinter.")



"""
Keep pygubu as a simple library. Translations will be done in pygubu-designer repo.

Here _ is used to mark strings in plugins files as translatable.

Do not use this file in your application projects.
"""

import os


def _real_translator(msg):
    return msg


_ = _real_translator


class LazyTranslator:
    def __init__(self, message):
        self._message = message

    def __str__(self):
        return _real_translator(self._message)

    def __repr__(self):
        return f"LazyTranslator('{self._message}')"


def setup_translator(translator):
    global _real_translator
    _real_translator = translator


def translator(message: str) -> str:
    return LazyTranslator(message)


def noop_translator(message: str) -> str:
    return message


if "PYGUBU_LAZY_TRANSLATOR" in os.environ:
    # Environment variable is set in pygubu-designer to activate string translations.
    _ = T = translator



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk


import fixpath
import pygubu
import support


class TestVariables(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_variables.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("mainwindow")

    def tearDown(self):
        support.root_withdraw()

    def test_string_var(self):
        var = self.builder.get_variable("strvar")
        self.assertIsInstance(var, tk.StringVar)
        self.widget.destroy()

    def test_int_var(self):
        var = self.builder.get_variable("intvar")
        self.assertIsInstance(var, tk.IntVar)
        self.widget.destroy()

    def test_double_var(self):
        var = self.builder.get_variable("doublevar")
        self.assertIsInstance(var, tk.DoubleVar)
        self.widget.destroy()

    def test_boolean_var(self):
        var = self.builder.get_variable("booleanvar")
        self.assertIsInstance(var, tk.BooleanVar)
        self.assertEqual(False, var.get())
        self.widget.destroy()

    def test_bugged_oldformat_var(self):
        var = self.builder.get_variable("testoldformat")
        self.assertIsInstance(var, tk.StringVar)
        self.widget.destroy()

    def test_invalid_variable_type(self):
        # self.builder.create_variable('complex:mycomplexvar')
        self.assertRaises(
            Exception, self.builder.create_variable, "complex:mycomplexvar"
        )
        self.widget.destroy()


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod


"""
Model transformers:

        transform(): "model data" => "norm data"
        reverseTransform(): "norm data" => "model data"

View transformers:

        transform(): "norm data" => "view data"
        reverseTransform(): "view data" => "norm data


 Model data     <|
 |               |
 | transform     | reverse transform
 |               |
 |>              |
    Norm data
 |              <|
 | transform     |
 |>              | reverse transform
    View data    |
"""


class DataTransformer(ABC):
    @abstractmethod
    def transform(self, value):
        ...

    @abstractmethod
    def reversetransform(self, value):
        ...


class NoopTransfomer(DataTransformer):
    def transform(self, value):
        return value

    def reversetransform(self, value):
        return value


class TransformationError(Exception):
    ...



import tkinter as tk
import tkinter.ttk as ttk
from collections import defaultdict

from pygubu.widgets.dockfw.slotframe import SlotFrame
from pygubu.widgets.dockfw.framework import (
    DockingFramework,
    IDockWidget,
    IDockFrame,
    IDockPane,
)


class DockWidgetBase(SlotFrame):
    def __init__(self, *args, maindock=None, uid=None, **kw):
        super().__init__(*args, **kw)
        self.maindock = maindock
        self.uid = uid
        self.parent_pane: DockPane = None
        self.dw_options = {}

    def child_master(self):
        return self.fcenter

    def detach_from_parent(self):
        """Remove this widget as child of parent_pane."""
        if self.parent_pane:
            if self in self.parent_pane.dw_children:
                self.parent_pane.dw_children.remove(self)
            # the panedwindow of parent could be deleted
            exists = self.tk.call(
                "winfo", "exists", str(self.parent_pane.panedw)
            )
            if exists:
                if self in self.parent_pane.panedw.panes():
                    self.parent_pane.panedw.remove(self)
            self.parent_pane = None


class DockPane(DockWidgetBase, IDockPane):
    pcount = 0

    def __init__(self, *args, orient=tk.HORIZONTAL, **kw):
        uid = kw.get("uid", None)
        if uid is None:
            type(self).pcount += 1
            kw["uid"] = f"pane{self.pcount}"
        super().__init__(*args, **kw)
        self.panedw = ttk.Panedwindow(self.fcenter, orient=orient)
        self.panedw.pack(expand=True, fill=tk.BOTH)
        self.dw_children = []

    @property
    def orient(self):
        return str(self.panedw.cget("orient"))

    @property
    def count(self):
        return len(self.panedw.panes())

    @property
    def sorted_dw_children(self):
        wsorted = []
        for tkpane in self.panedw.panes():
            tkpane = self.nametowidget(tkpane)
            if isinstance(tkpane, ttk.Notebook):
                for tab in tkpane.tabs():
                    tab = self.nametowidget(tab)
                    if tab in self.dw_children:
                        wsorted.append(tab)
            else:
                if tkpane in self.dw_children:
                    wsorted.append(tkpane)
        return wsorted

    def add_pane(self, pane: "DockPane", **pane_kw):
        pane.dw_options.update(pane_kw)
        self.maindock._add_pane_to_pane(self, pane, **pane_kw)

    def add_widget(self, widget: "DockWidget", grouped=False, weight=1):
        widget.dw_options["weight"] = weight
        self.maindock._add_widget_to_pane(
            self, widget, grouped=grouped, weight=weight
        )

    def add_dwchild(self, dw: DockWidgetBase):
        """Add DockPane or DockWidget as child."""
        dw.parent_pane = self
        self.dw_children.append(dw)


class DockWidget(DockWidgetBase, IDockWidget):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.noteb: ttk.Notebook = None
        self._title = None

    @property
    def is_grouped(self):
        grouped = False
        if self.noteb is not None:
            if len(self.noteb.tabs()) > 1:
                grouped = True
        return grouped

    @property
    def title(self):
        return self._title if self._title is not None else self.uid

    def configure(self, cnf=None, **kw):
        if cnf:
            return super().configure(cnf, **kw)
        key = "title"
        if key in kw:
            self._title = kw.pop(key)
        return super().configure(cnf, **kw)



import tkinter as tk
import tkinter.ttk as ttk
from abc import ABC, ABCMeta, abstractmethod


class IThemeBuilder(ABC):
    @abstractmethod
    def theme_name(self) -> str:
        ...

    @abstractmethod
    def theme_parent(self) -> str:
        ...

    @abstractmethod
    def theme_settings(self):
        ...

    @abstractmethod
    def db_settings(self):
        ...

    @abstractmethod
    def tk_palette(self):
        ...

    def create(self, master):
        style = ttk.Style(master)
        name = self.theme_name()
        parent = self.theme_parent()
        if name not in style.theme_names():
            theme_settings = self.theme_settings()
            style.theme_create(name, parent, theme_settings)
        else:
            raise ValueError("Theme name already exists.")

    def apply(self, master: tk.Widget):
        # master.option_clear()

        palette = self.tk_palette()
        if palette:
            master.tk_setPalette(**palette)

        db_settings = self.db_settings()
        for pattern, options in db_settings.items():
            for option, value in options.items():
                fpattern = f"{pattern}{option}"
                master.option_add(fpattern, value, "widgetDefault")
        style = ttk.Style(master)
        style.theme_use(self.theme_name())



import tkinter as tk
from . import DataTransformer


class BoolTransformer(DataTransformer):
    def transform(self, value):
        tval = False
        try:
            tval = tk.getboolean(value)
        except (ValueError, tk.TclError):
            pass
        return tval

    def reversetransform(self, value):
        return self.transform(value)



import tkinter as tk
from pygubu.api.v1 import BuilderObject
from pygubu.plugins.tk.tkstdwidgets import TKToplevel


class ToplevelPreviewFactory(type):
    def __new__(cls, clsname, superclasses, attrs):
        return type.__new__(cls, str(clsname), superclasses, attrs)


class ToplevelPreviewMixin(object):
    def __init__(self, master=None, **kw):
        super().__init__(master, **kw)
        self.tl_attrs = {}
        self._w_set = False  # Marks if width was set using "width" property
        self._h_set = False  # Marks if height was set using "height" property
        self._propagate = True  # Save propagate state configured by user
        self._geometry_set = False
        self._geom_w = 1
        self._geom_h = 1
        self._uwidth = 1  # Save width configured by user
        self._uheight = 1  # Save height configured by user

    def _get_req_wh(self):
        """Calculate aproximation of real width and height
        because this frame is locked for preview purposes."""
        my_w = self._uwidth
        my_h = self._uheight
        has_minsize = False
        min_w = my_w
        min_h = my_h
        final_w = my_w
        final_h = my_h

        if "minsize" in self.tl_attrs:
            minsize = self.tl_attrs.get("minsize", (1, 1))
            min_w = minsize[0]
            min_h = minsize[1]
            has_minsize = True

        content_w = content_h = 0
        if self._propagate:
            content_w = super().winfo_reqwidth()
            content_h = super().winfo_reqheight()
            children = self.winfo_children()
            if children:
                if has_minsize:
                    # propagate is enabled, resize to
                    # content size or minsize
                    final_w = max(content_w, min_w)
                    final_h = max(content_h, min_h)
                else:
                    final_w = content_w
                    final_h = content_h
            else:
                # No children, show frame with user configured size
                final_w = max(my_w, min_w)
                final_h = max(my_h, min_h)
        else:
            # Propagate is false, fixed w/h frame.
            final_w = max(min_w, my_w)
            final_h = max(min_h, my_h)
        if self._geometry_set:
            # Geometry overrides all
            final_w = self._geom_w
            final_h = self._geom_h
        # print(f"{my_w=}, {my_h=}, {has_minsize=}, {min_w=}, {min_h=}, {content_w=}, {content_h=}")
        # print(f"{final_w=}, {final_h=}")
        return (final_w, final_h)

    def winfo_reqwidth(self):
        return self._get_req_wh()[0]

    def winfo_reqheight(self):
        return self._get_req_wh()[1]

    def configure(self, cnf=None, **kw):
        if cnf:
            return super().configure(cnf, **kw)
        key = "width"
        if key in kw:
            value = int(kw[key])
            minsize = self.tl_attrs.get("minsize", None)
            maxsize = self.tl_attrs.get("maxsize", None)
            remove = False
            if minsize and value < minsize[0]:
                remove = True
            if maxsize and value > maxsize[0]:
                remove = True
            if self._w_set:
                resizable = self.tl_attrs.get("resizable", None)
                if resizable and not TKToplevel.RESIZABLE[resizable][0]:
                    remove = True
            if remove:
                kw.pop(key)
            else:
                self._w_set = True
            # save user width setting
            self._uwidth = value
        key = "height"
        if key in kw:
            value = int(kw[key])
            minsize = self.tl_attrs.get("minsize", None)
            maxsize = self.tl_attrs.get("maxsize", None)
            remove = False
            if minsize and value < minsize[1]:
                remove = True
            if maxsize and value > maxsize[1]:
                remove = True
            if self._h_set:
                resizable = self.tl_attrs.get("resizable", None)
                if resizable and not TKToplevel.RESIZABLE[resizable][1]:
                    remove = True
            if remove:
                kw.pop(key)
            else:
                self._h_set = True
            # save user height setting
            self._uheight = value
        key = "menu"
        if key in kw:
            # No menu preview available
            kw.pop(key)
        return super().configure(cnf, **kw)


class ToplevelPreviewBaseBO(BuilderObject):
    container = True
    container_layout = True
    # Add fake 'modal' property for Dialog preview
    properties = TKToplevel.properties + ("modal",)
    ro_properties = TKToplevel.ro_properties

    def configure(self, target=None):
        # setup width and height properties if
        # geometry is defined.
        geom = "geometry"
        if geom in self.wmeta.properties:
            w, h = self._get_dimwh(self.wmeta.properties[geom])
            if w and h:
                self.wmeta.properties["width"] = w
                self.wmeta.properties["height"] = h
        super().configure(target)

    def configure_children(self, target=None):
        w = self.widget
        # Set a fixed size in preview if geometry is set.
        if "geometry" in self.wmeta.properties:
            if w.pack_slaves():
                w.pack_propagate(0)
            elif w.grid_slaves():
                w.grid_propagate(0)

    def _container_layout(self, target, container_manager, properties):
        # save user configured propagate state
        value = self.wmeta.container_properties.get("propagate", True)
        value = self._process_property_value("propagate", value)
        self.widget._propagate = value
        super()._container_layout(target, container_manager, properties)

    def layout(self, target=None):
        self.widget.pack(expand=True, fill="both")
        self._container_layout(
            self.widget,
            self.wmeta.container_manager,
            self.wmeta.container_properties,
        )

    def _get_dimwh(self, dimvalue: str):
        # get width and height from dimension string
        dim = dimvalue.split("+")[0]
        dim = dim.split("-")[0]
        w, h = dim.split("x")
        return (int(w), int(h))

    def _process_property_value(self, pname, value):
        if pname in ("maxsize", "minsize"):
            if "|" in value:
                w, h = value.split("|")
                value = (int(w), int(h))
            return value
        if pname == "propagate":
            return tk.getboolean(value)
        return super()._process_property_value(pname, value)

    def _set_property(self, target_widget, pname, value):
        tw = target_widget
        tw.tl_attrs[pname] = value
        method_props = (
            "iconbitmap",
            "iconphoto",
            "overrideredirect",
            "title",
        )
        if pname in method_props:
            pass
        elif pname in ("maxsize", "minsize"):
            pvalue = self._process_property_value(pname, value)
            if not pvalue:
                del tw.tl_attrs[pname]
            else:
                if isinstance(pvalue, tuple):
                    tw.tl_attrs[pname] = pvalue
                    if pname == "minsize":
                        tw.configure(width=pvalue[0], height=pvalue[1])
                else:
                    del tw.tl_attrs[pname]
        elif pname == "geometry":
            if value:
                w, h = self._get_dimwh(value)
                if w and h:
                    w, h = int(w), int(h)
                    tw.tl_attrs["minsize"] = (w, h)
                    tw._h_set = tw._w_set = False
                    tw.configure(width=w, height=h)
                    tw._geometry_set = True
                    tw._geom_w = w
                    tw._geom_h = h
        elif pname == "resizable":
            # Do nothing, fake 'resizable' property for Toplevel preview
            pass
        elif pname == "modal":
            # Do nothing, fake 'modal' property for dialog preview
            pass
        else:
            super()._set_property(tw, pname, value)



import awesometkinter as atk
from pygubu.i18n import _
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)

from ..awesometkinter import _designer_tab_label, _plugin_uid


class RadialProgressbarBO(BuilderObject):
    OPTIONS_CUSTOM = (
        "bg",
        "fg",
        "text_fg",
        "font",
        "font_size_ratio",
        "base_img",
        "indicator_img",
        "parent_bg",
    )
    ro_properties = OPTIONS_CUSTOM
    class_ = atk.RadialProgressbar

    def _process_property_value(self, pname, value):
        if pname == "font_size_ratio":
            value_ = 0.1
            try:
                value_ = float(value)
            except ValueError:
                pass
            return value_
        return super()._process_property_value(pname, value)


_builder_uid = _plugin_uid + ".RadialProgressbar"
register_widget(
    _builder_uid,
    RadialProgressbarBO,
    "RadialProgressbar",
    ("ttk", _designer_tab_label),
)
register_custom_property(
    _builder_uid,
    "bg",
    "colorentry",
    help=_("color of base ring"),
)
register_custom_property(
    _builder_uid,
    "fg",
    "colorentry",
    help=_("color of indicator ring"),
)
register_custom_property(
    _builder_uid,
    "text_fg",
    "colorentry",
    help=_("percentage text color"),
)
register_custom_property(
    _builder_uid,
    "font",
    "fontentry",
    help=_("tkinter font for percentage text"),
)
register_custom_property(
    _builder_uid,
    "font_size_ratio",
    "spinbox",
    from_=0.1,
    to=1,
    increment=0.1,
    help=_(
        "font size to progressbar width ratio, e.g. for a progressbar size 100 pixels, a 0.1 ratio means font size 10"
    ),
)
register_custom_property(
    _builder_uid,
    "base_img",
    "imageentry",
    help=_("base image for progressbar"),
)
register_custom_property(
    _builder_uid,
    "indicator_img",
    "imageentry",
    help=_("indicator image for progressbar"),
)
register_custom_property(
    _builder_uid,
    "parent_bg",
    "colorentry",
    help=_("color of parent container"),
)


class RadialProgressbar3dBO(BuilderObject):
    OPTIONS_CUSTOM = ("fg", "text_fg", "text_bg")
    ro_properties = OPTIONS_CUSTOM
    class_ = atk.RadialProgressbar3d


_builder_uid = _plugin_uid + ".RadialProgressbar3d"
register_widget(
    _builder_uid,
    RadialProgressbar3dBO,
    "RadialProgressbar3d",
    ("ttk", _designer_tab_label),
)
register_custom_property(
    _builder_uid,
    "fg",
    "colorentry",
)
register_custom_property(
    _builder_uid,
    "text_fg",
    "colorentry",
)
register_custom_property(
    _builder_uid,
    "text_bg",
    "colorentry",
)


class SegmentbarBO(BuilderObject):
    OPTIONS_CUSTOM = ("bg", "fg", "width", "height")
    ro_properties = OPTIONS_CUSTOM
    class_ = atk.Segmentbar

    def _process_property_value(self, pname, value):
        if pname in ("width",):
            return int(value)
        return super()._process_property_value(pname, value)


_builder_uid = _plugin_uid + ".Segmentbar"
register_widget(
    _builder_uid, SegmentbarBO, "Segmentbar", ("ttk", _designer_tab_label)
)
register_custom_property(
    _builder_uid,
    "bg",
    "colorentry",
)
register_custom_property(
    _builder_uid,
    "fg",
    "colorentry",
)
register_custom_property(
    _builder_uid,
    "width",
    "naturalnumber",
)
register_custom_property(
    _builder_uid,
    "height",
    "dimensionentry",
)



""" Widget utility functions."""
import tkinter as tk


def crop_widget(widget, *, recursive=False):
    """Remove standard widget functionality."""

    # Remove standard bindings
    wclass = widget.winfo_class()
    bindtags = widget.bindtags()
    if wclass in bindtags:
        bindtags = list(bindtags)
        bindtags.remove(wclass)
        widget.bindtags(bindtags)
    # Don't take focus
    try:
        widget.configure(takefocus=False)
    except Exception:
        pass

    if recursive:
        for childw in widget.winfo_children():
            crop_widget(childw, recursive=recursive)


def iter_parents(widget: tk.Widget):
    top = str(widget.winfo_toplevel())
    parent = widget.winfo_parent()
    if not parent or parent == ".":
        return None
    while top != parent:
        pw = widget.nametowidget(parent)
        yield pw
        parent = pw.winfo_parent()


def iter_to_toplevel(widget: tk.Widget):
    """Iter parents of widget including widget itself"""
    yield widget
    yield from iter_parents(widget)


class HideableMixin:
    """Allows to hide a tk.Widget."""

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._is_hidden = None
        self._layout = None
        self._layout_info = None

    @property
    def hidden(self):
        return self._is_hidden

    @hidden.setter
    def hidden(self, state: bool):
        if state:
            self._hide()
        else:
            self._show()

    def _hide(self):
        if self._is_hidden is None:
            self._layout = m = self.winfo_manager()
            if m == "pack":
                self._layout_info = self.pack_info()
                self._layout_info.pop("in", None)
                parent = self.nametowidget(self.winfo_parent())
                wlist = parent.pack_slaves()
                total = len(wlist)
                self_index = wlist.index(self)
                if total > 1:
                    if self_index == 0:
                        self._layout_info["before"] = wlist[1]
                    else:
                        self._layout_info["after"] = wlist[self_index - 1]

            elif m == "place":
                self._layout_info = info = self.place_info()
                info.pop("in", None)
                # FIXME: ttkbootstrap issue with localization ??
                for key in ("relx", "rely", "relwidth", "relheight"):
                    info[key] = str(info[key]).replace(",", ".")
            self._is_hidden = False

        if self._is_hidden is False:
            if self._layout == "pack":
                self.pack_forget()
            elif self._layout == "grid":
                self.grid_remove()
            elif self._layout == "place":
                self.place_forget()
            self._is_hidden = True

    def _show(self):
        if self._is_hidden:
            layout = self._layout
            if layout == "pack":
                self.pack(**self._layout_info)
            elif layout == "grid":
                self.grid()
            elif layout == "place":
                self.place(**self._layout_info)
            self._is_hidden = False



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk

import fixpath
import support
import pygubu


class TestUIDefinition_1_1(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_uidefinition_1_1.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("fmain")
        self.fgrid = builder.get_object("fgrid")

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.assertIsInstance(self.widget, ttk.Frame)
        self.widget.destroy()

    def test_gridrc_options(self):
        config = self.fgrid.grid_rowconfigure(0)
        # {'minsize': 10, 'pad': 10, 'weight': 1, 'uniform': 'x'}
        self.assertEqual(config["minsize"], 10)
        self.assertEqual(config["pad"], 10)
        self.assertEqual(config["weight"], 1)
        self.assertEqual(config["uniform"], "x")

        config = self.fgrid.grid_columnconfigure(0)
        # {'minsize': 30, 'pad': 15, 'weight': 1, 'uniform': 'y'}
        self.assertEqual(config["minsize"], 30)
        self.assertEqual(config["pad"], 15)
        self.assertEqual(config["weight"], 1)
        self.assertEqual(config["uniform"], "y")



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk


import fixpath
import pygubu
import support

import pygubu.widgets.scrollbarhelper
import pygubu.widgets.tkscrollbarhelper


class TestScrollbarHelper(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = """<?xml version="1.0" ?>
<interface>
  <object class="pygubu.builder.widgets.scrollbarhelper" id="scrollbarhelper">
    <property name="scrolltype">both</property>
    <property name="padding">5 5 5 5</property>
    <property name="relief">raised</property>
    <layout>
      <property name="column">0</property>
      <property name="propagate">True</property>
      <property name="row">0</property>
    </layout>
    <child>
      <object class="tk.Canvas" id="tk.Canvas_1">
        <property name="background">#d9d900</property>
        <property name="scrollregion">0 0 10i 10i</property>
        <layout>
          <property name="column">0</property>
          <property name="propagate">True</property>
          <property name="row">0</property>
        </layout>
      </object>
    </child>
  </object>
</interface>
"""
        builder = pygubu.Builder()
        builder.add_from_string(xmldata)
        self.widget = builder.get_object("scrollbarhelper")

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.assertIsInstance(
            self.widget, pygubu.widgets.scrollbarhelper.ScrollbarHelper
        )
        self.widget.destroy()

    def test_padding(self):
        expected_value = ("5", "5", "5", "5")
        tclobj = self.widget.cget("padding")
        padding = (
            str(tclobj[0]),
            str(tclobj[1]),
            str(tclobj[2]),
            str(tclobj[3]),
        )
        self.assertEqual(expected_value, padding)
        self.widget.destroy()


class TestTkScrollbarHelper(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = """<?xml version="1.0" ?>
<interface>
  <object class="pygubu.builder.widgets.tkscrollbarhelper" id="scrollbarhelper">
    <property name="scrolltype">both</property>
    <property name="padx">5</property>
    <property name="pady">10</property>
    <property name="relief">raised</property>
    <layout>
      <property name="column">0</property>
      <property name="propagate">True</property>
      <property name="row">0</property>
    </layout>
    <child>
      <object class="tk.Canvas" id="tk.Canvas_1">
        <property name="background">#d9d900</property>
        <property name="scrollregion">0 0 10i 10i</property>
        <layout>
          <property name="column">0</property>
          <property name="propagate">True</property>
          <property name="row">0</property>
        </layout>
      </object>
    </child>
  </object>
</interface>
"""
        builder = pygubu.Builder()
        builder.add_from_string(xmldata)
        self.widget = builder.get_object("scrollbarhelper")

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.assertIsInstance(
            self.widget, pygubu.widgets.tkscrollbarhelper.TkScrollbarHelper
        )
        self.widget.destroy()

    def test_padx(self):
        prop = "padx"
        expected_value = "5"
        tclobj = self.widget.cget(prop)
        value = str(tclobj)
        self.assertEqual(expected_value, value)
        self.widget.destroy()

    def test_pady(self):
        prop = "pady"
        expected_value = "10"
        tclobj = self.widget.cget(prop)
        value = str(tclobj)
        self.assertEqual(expected_value, value)
        self.widget.destroy()


if __name__ == "__main__":
    unittest.main()



# encoding: utf-8
import tkinter as tk
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.i18n import _
from pygubu.plugins.ttk.ttkstdwidgets import TTKButton
from pygubu.widgets.pathchooserinput import PathChooserInput, PathChooserButton


class PathChooserBaseMixin:
    base_properties = (
        "type",
        "path",
        "initialdir",
        "mustexist",
        "title",
        "defaultextension",
    )

    def _process_property_value(self, pname, value):
        if pname == "mustexist":
            return tk.getboolean(value)
        return super()._process_property_value(pname, value)

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname == "mustexist":
            return self._process_property_value(pname, value)
        return super()._code_process_property_value(targetid, pname, value)


# common properties
_builder_id = "pygubu.widgets.PathChooser.*"
_help = _("Dialog type")
register_custom_property(
    _builder_id,
    "type",
    "choice",
    values=(PathChooserInput.FILE, PathChooserInput.DIR),
    state="readonly",
    default_value=PathChooserInput.FILE,
    help=_help,
)

_help = _("Initial path value.")
register_custom_property(_builder_id, "path", "entry", help=_help)

_help = _(
    "Dialog option. Determines if path must exist for directory and file dialogs. The default value is True."
)
register_custom_property(
    _builder_id,
    "mustexist",
    "choice",
    values=("", "true", "false"),
    state="readonly",
    default_value="true",
    help=_help,
)

_help = _("Dialog option. Sets initial directory.")
register_custom_property(_builder_id, "initialdir", "entry", help=_help)

_help = _("Dialog option. Sets dialog title.")
register_custom_property(_builder_id, "title", "entry", help=_help)

_help = _("Dialog option. Sets default file extension.")
register_custom_property(_builder_id, "defaultextension", "entry", help=_help)


class PathChooserInputBO(PathChooserBaseMixin, BuilderObject):
    class_ = PathChooserInput
    properties = PathChooserBaseMixin.base_properties + (
        "image",
        "textvariable",
        "state",
    )
    virtual_events = ("<<PathChooserPathChanged>>",)


_builder_id = "pygubu.widgets.PathChooserInput"
register_widget(
    _builder_id,
    PathChooserInputBO,
    "PathChooserInput",
    ("ttk", _("Pygubu Widgets")),
)
_old_bid = "pygubu.builder.widgets.pathchooserinput"
register_widget(
    _old_bid,
    PathChooserInputBO,
    public=False,
)

_help = _("Image for the button.")
register_custom_property(_builder_id, "image", "imageentry", help=_help)

_help = _("Tk variable associated to the path property.")
register_custom_property(_builder_id, "textvariable", "tkvarentry", help=_help)

_help = _("Path entry state.")
register_custom_property(
    _builder_id,
    "state",
    "choice",
    values=("", "normal", "disabled", "readonly"),
    state="readonly",
    help=_help,
)


class PathChooserButtonBO(PathChooserBaseMixin, BuilderObject):
    class_ = PathChooserButton
    properties = PathChooserBaseMixin.base_properties + tuple(
        set(TTKButton.properties) - set(("command", "default"))
    )
    virtual_events = ("<<PathChooserPathChanged>>",)


_builder_id = "pygubu.widgets.PathChooserButton"
register_widget(
    _builder_id,
    PathChooserButtonBO,
    "PathChooserButton",
    ("ttk", _("Pygubu Widgets")),
)

register_custom_property(_builder_id, "width", "integernumber")



from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.ttk.ttkstdwidgets import TTKFrame
from pygubu.i18n import _

from pygubu.widgets.hideableframe import HideableFrame


class HideableFrameBO(TTKFrame):
    class_ = HideableFrame


_builder_uid = "pygubu.widgets.hideableframe"
register_widget(
    _builder_uid,
    HideableFrameBO,
    "HideableFrame",
    (_("Pygubu Widgets"), "ttk"),
)

register_custom_property(
    _builder_uid, "width", "dimensionentry", default_value=200
)

register_custom_property(
    _builder_uid, "height", "dimensionentry", default_value=200
)



import os
from pygubu.api.v1 import BuilderLoaderPlugin


class PygubuWidgetsLoader(BuilderLoaderPlugin):
    module_map = {
        "pygubu.plugins.pygubu.calendarframe": (
            "pygubu.builder.widgets.calendarframe",
        ),
        "pygubu.plugins.pygubu.combobox": ("pygubu.builder.widgets.combobox",),
        "pygubu.plugins.pygubu.dialog": ("pygubu.builder.widgets.dialog",),
        "pygubu.plugins.pygubu.editabletreeview": (
            "pygubu.builder.widgets.editabletreeview",
        ),
        "pygubu.plugins.pygubu.pathchooserinput": (
            "pygubu.builder.widgets.pathchooserinput",
            "pygubu.widgets.PathChooserInput",
            "pygubu.widgets.PathChooserButton",
        ),
        "pygubu.plugins.pygubu.scrollbarhelper": (
            "pygubu.builder.widgets.scrollbarhelper",
        ),
        "pygubu.plugins.pygubu.scrolledframe": (
            "pygubu.builder.widgets.scrolledframe",
        ),
        "pygubu.plugins.pygubu.tkinterscrolledtext": (
            "pygubu.builder.widgets.tkinterscrolledtext",
        ),
        "pygubu.plugins.pygubu.tkscrollbarhelper": (
            "pygubu.builder.widgets.tkscrollbarhelper",
        ),
        "pygubu.plugins.pygubu.tkscrolledframe": (
            "pygubu.builder.widgets.tkscrolledframe",
        ),
        "pygubu.plugins.pygubu.filterabletreeview": (
            "pygubu.widgets.FilterableTreeview",
        ),
        "pygubu.plugins.pygubu.accordionframe": (
            "pygubu.widgets.AccordionFrame",
        ),
        "pygubu.plugins.pygubu.dockfw": (
            "pygubu.widgets.dockframe",
            "pygubu.widgets.dockpane",
            "pygubu.widgets.dockwidget",
        ),
        "pygubu.plugins.pygubu.hideableframe": ("pygubu.widgets.hideableframe"),
        "pygubu.plugins.pygubu.colorinputbo": ("pygubu.widgets.ColorInput"),
        # Forms are not finished so expect changes
        "pygubu.plugins.pygubu.forms.ttkwidgetbo": (
            "pygubu.forms.ttkwidget.FrameFormBuilder",
            "pygubu.forms.ttkwidget.Label",
            "pygubu.forms.ttkwidget.Entry",
            "pygubu.forms.ttkwidget.LabelWidgetInfo",
        ),
        "pygubu.plugins.pygubu.forms.tkwidgetbo": (
            "pygubu.forms.tkwidget.Text",
        ),
        "pygubu.plugins.pygubu.forms.pygubuwidgetbo": (
            "pygubu.forms.pygubuwidget.PygubuCombobox"
        ),
    }

    def do_activate(self) -> bool:
        return True

    def get_all_modules(self):
        return [m for m in self.module_map.keys()]

    def get_module_for(self, identifier: str) -> str:
        for module, identifiers in self.module_map.items():
            if identifier in identifiers:
                return module
        return None

    def can_load(self, identifier: str) -> bool:
        for module, builders in self.module_map.items():
            if identifier in builders:
                return True
        return False

    def get_designer_plugin(self):
        """Load class that implements IDesignerPlugin"""
        from .designer.designerplugin import PygubuDesignerPlugin

        return PygubuDesignerPlugin()



import tkinter.ttk as ttk
from pygubu.api.v1 import BuilderObject, register_widget


class CustomLabel(ttk.Label):
    def get_message(self):
        return "CustomLabel"


class TestCustomWidgetBuilder(BuilderObject):
    class_ = CustomLabel


register_widget(
    "test_custom_widget_module.custom_label",
    TestCustomWidgetBuilder,
    "CustomWidget",
    ("ttk", "Test Custom Widget"),
)



import functools
import tkinter as tk
import tkinter.ttk as ttk
from abc import ABC, abstractmethod
from collections import defaultdict
from enum import Enum


class InplaceEditor(ABC):
    event_value_changed = "<<InplaceEditor:ValueChanged>>"

    def _notify_change(self) -> None:
        self.widget.event_generate(self.event_value_changed)

    def focus_set(self) -> None:
        self.widget.focus_set()

    @property
    @abstractmethod
    def widget(self) -> tk.Widget:
        ...

    @property
    @abstractmethod
    def value(self):
        ...

    @abstractmethod
    def edit(self, value) -> None:
        ...


class _VariableBasedEditor(InplaceEditor):
    @abstractmethod
    def _create_widget(self, master, **kw) -> tk.Widget:
        ...

    def __init__(self, master, **kw):
        self._var_blocked = False
        self._variable = None
        if kw is None:
            kw = {}
        if "textvariable" in kw:
            self._variable = kw["textvariable"]
        else:
            self._variable = tk.StringVar()
            kw["textvariable"] = self._variable
        self._widget = self._create_widget(master, **kw)

        def on_var_write(var, index, mode):
            if not self._var_blocked:
                self._notify_change()

        self._variable.trace_add("write", on_var_write)

    @property
    def widget(self):
        return self._widget

    @property
    def value(self):
        return self._variable.get()

    def edit(self, value):
        self._var_blocked = True
        self._variable.set(value)
        self._var_blocked = False


class _EntryEditor(_VariableBasedEditor):
    def _create_widget(self, master, **kw) -> tk.Widget:
        return ttk.Entry(master, **kw)


class _CheckbuttonEditor(_VariableBasedEditor):
    def _create_widget(self, master, **kw) -> tk.Widget:
        return ttk.Checkbutton(master, **kw)


class _ComboboxEditor(_VariableBasedEditor):
    def _create_widget(self, master, **kw) -> tk.Widget:
        return ttk.Combobox(master, **kw)


class _SpinboxEditor(_VariableBasedEditor):
    def _create_widget(self, master, **kw) -> tk.Widget:
        return ttk.Spinbox(master, **kw)


class _CustomEditor(_VariableBasedEditor):
    def __init__(self, widget, **kw):
        self._widget = widget
        super().__init__(widget, **kw)

    def _create_widget(self, master, **kw) -> tk.Widget:
        return self._widget


class _EditorType(Enum):
    ENTRY = 1
    CHECKBUTTON = 2
    COMBOBOX = 3
    SPINBOX = 4
    CUSTOM_WIDGET = 5
    CUSTOM_EDITOR = 6


class EditableTreeview(ttk.Treeview):
    """A simple editable treeview

    It uses the following events from Treeview:
        <<TreviewSelect>>
        <4>
        <5>
        <KeyRelease>
        <Home>
        <End>
        <Configure>
        <Button-1>
        <ButtonRelease-1>
        <Motion>
    If you need them use add=True when calling bind method.

    It Generates three virtual events:
        <<TreeviewInplaceEdit>>
        <<TreeviewCellEdited>>
        <<TreeviewEditorsUnfocused>>

    <<TreeviewInplaceEdit>> is emitted and used to configure cell editors.
    <<TreeviewCellEdited>> is emitted after a cell was changed.
    <<TreeviewEditorsUnfocused>> is emitted when user clicks in treeview
    white area where no rows are rendered.

    You can know wich cell is being configured or edited, using:
        get_event_info()

    To quickly get data from tree columns you can use:
        get_value(col, item)

    """

    def __init__(self, master=None, **kw):
        super().__init__(master, **kw)

        self._curfocus = None
        self._editors = {}
        self._editors_show = {}
        self._editors_bag = defaultdict(dict)  # Multiple editors per row
        self._header_clicked = False
        self._header_dragged = False
        self._last_column_clicked = "#0"
        self._update_callback_id = None

        self.bind("<<TreeviewSelect>>", self.__check_focus)
        # Wheel events?
        self.bind("<4>", self._schedule_update)
        self.bind("<5>", self._schedule_update)
        self.bind("<KeyRelease>", self.__check_focus)
        self.bind("<Home>", functools.partial(self.__on_key_press, "Home"))
        self.bind("<End>", functools.partial(self.__on_key_press, "End"))
        self.bind("<Button-1>", self.__on_button1)
        self.bind("<ButtonRelease-1>", self.__on_button1_release)
        self.bind("<Motion>", self.__on_mouse_motion)
        self.bind("<Configure>", self._schedule_update)

    def _schedule_update(self, event=None):
        if self._update_callback_id is None:
            self._update_callback_id = self.after_idle(self.__updateWnds)

    def __on_button1(self, event):
        r = self.identify_region(event.x, event.y)
        if r in ("separator", "header"):
            self._header_clicked = True
        elif r in ("tree", "cell"):
            if not self._editors_show:
                self._schedule_update()
        elif r == "nothing":
            if self._editors_show:
                self.__clear_inplace_widgets()
                self._curfocus = None
                self.event_generate("<<TreeviewEditorsUnfocused>>")
        self._last_column_clicked = self.identify_column(event.x)

    def __on_mouse_motion(self, event):
        if self._header_clicked:
            self._header_dragged = True

    def __on_button1_release(self, event):
        if self._header_dragged:
            self._schedule_update(event)
        self._header_clicked = False
        self._header_dragged = False

    def __on_key_press(self, key, event):
        if key == "Home":
            self.selection_set("")
            self.focus(self.get_children()[0])
        if key == "End":
            self.selection_set("")
            self.focus(self.get_children()[-1])

    def delete(self, *items):
        self._schedule_update()
        ttk.Treeview.delete(self, *items)

    def yview(self, *args):
        """Update inplace widgets position when doing vertical scroll"""
        self._schedule_update()
        ttk.Treeview.yview(self, *args)

    def yview_scroll(self, number, what):
        self._schedule_update()
        ttk.Treeview.yview_scroll(self, number, what)

    def yview_moveto(self, fraction):
        self._schedule_update()
        ttk.Treeview.yview_moveto(self, fraction)

    def xview(self, *args):
        """Update inplace widgets position when doing horizontal scroll"""
        self._schedule_update()
        ttk.Treeview.xview(self, *args)

    def xview_scroll(self, number, what):
        self._schedule_update()
        ttk.Treeview.xview_scroll(self, number, what)

    def xview_moveto(self, fraction):
        self._schedule_update()
        ttk.Treeview.xview_moveto(self, fraction)

    def __check_focus(self, event):
        """Checks if the focus has changed"""
        changed = False
        if not self._curfocus:
            changed = True
        elif self._curfocus != self.focus():
            self.__clear_inplace_widgets()
            changed = True
        newfocus = self.focus()
        if changed:
            if newfocus:
                # print('Focus changed to:', newfocus)
                self._curfocus = newfocus
                self.__focus(newfocus)
            self.__updateWnds()

    def __focus(self, item):
        """Called when focus item has changed"""
        cols = self.__get_display_columns()
        for col in cols:
            self.__event_info = (col, item)
            self.event_generate("<<TreeviewInplaceEdit>>")

    def __updateWnds(self, event=None):
        if not self._curfocus:
            for col, editor in self._editors.items():
                editor.widget.place_forget()
            self._update_callback_id = None
            return
        item = self._curfocus
        item_exists = self.exists(item)
        cols = self.__get_display_columns()
        col_diff = 0 if "#0" in cols else -1
        last_column_index = int(self._last_column_clicked[1:]) + col_diff
        for index, col in enumerate(cols):
            if col in self._editors:
                editor = self._editors[col]
                bbox = "" if not item_exists else self.bbox(item, column=col)
                if bbox == "":
                    editor.widget.place_forget()
                elif col in self._editors_show:
                    editor.widget.place(
                        x=bbox[0], y=bbox[1], width=bbox[2], height=bbox[3]
                    )
                    # try to focus the widget in the column clicked
                    if last_column_index == index:
                        editor.focus_set()
        self._update_callback_id = None

    def __clear_inplace_widgets(self):
        """Remove all inplace edit widgets."""
        for col, editor in self._editors.items():
            editor.widget.place_forget()
        self._editors_show.clear()

    def __get_display_columns(self):
        cols = self.cget("displaycolumns")
        show = (str(s) for s in self.cget("show"))
        if "#all" in cols or "tree" in show:
            cols = ("#0",) + self.cget("columns")
        return cols

    def get_event_info(self):
        return self.__event_info

    def get_value(self, col, item):
        """Return data value of tree item at the specified column index."""
        return self.__get_value(col, item)

    # FIXME:
    # def hide_editors(self):
    #    self.__clear_inplace_widgets()
    #

    def __get_value(self, col, item):
        if col == "#0":
            return self.item(item, "text")
        else:
            return self.set(item, col)

    def __set_value(self, col, item, value):
        if col == "#0":
            self.item(item, text=value)
        else:
            self.set(item, col, value)
        self.__event_info = (col, item)
        self.event_generate("<<TreeviewCellEdited>>")

    def __update_value(self, col, item):
        if not self.exists(item):
            return
        value = self.__get_value(col, item)
        newvalue = self._editors[col].value
        if value != newvalue:
            self.__set_value(col, item, newvalue)

    def _setup_editor(self, col, item, editor):
        editor.edit(self.__get_value(col, item))

        def on_value_change(event):
            self.__update_value(col, item)

        editor.widget.bind(InplaceEditor.event_value_changed, on_value_change)
        self._editors_show[col] = True

    def inplace_entry(self, col, item):
        current_editor = None
        if _EditorType.ENTRY not in self._editors_bag[col]:
            current_editor = _EntryEditor(self)
            self._editors_bag[col][_EditorType.ENTRY] = current_editor
        else:
            current_editor = self._editors_bag[col][_EditorType.ENTRY]
        self._editors[col] = current_editor
        self._setup_editor(col, item, current_editor)

    def inplace_checkbutton(self, col, item, onvalue="True", offvalue="False"):
        current_editor = None
        if _EditorType.CHECKBUTTON not in self._editors_bag[col]:
            svar = tk.StringVar()
            current_editor = _CheckbuttonEditor(
                self,
                textvariable=svar,
                variable=svar,
                onvalue=onvalue,
                offvalue=offvalue,
            )
            self._editors_bag[col][_EditorType.CHECKBUTTON] = current_editor
        else:
            current_editor = self._editors_bag[col][_EditorType.CHECKBUTTON]
        self._editors[col] = current_editor
        self._setup_editor(col, item, current_editor)

    def inplace_combobox(
        self, col, item, values, readonly=True, update_values=False
    ):
        current_editor = None
        if _EditorType.COMBOBOX not in self._editors_bag[col]:
            state = "readonly" if readonly else "normal"
            current_editor = _ComboboxEditor(self, values=values, state=state)
            self._editors_bag[col][_EditorType.COMBOBOX] = current_editor
        else:
            current_editor = self._editors_bag[col][_EditorType.COMBOBOX]
        self._editors[col] = current_editor
        self._setup_editor(col, item, current_editor)
        if update_values:
            current_editor.widget.configure(values=values)

    def inplace_spinbox(self, col, item, min, max, step):
        current_editor = None
        if _EditorType.SPINBOX not in self._editors_bag[col]:
            current_editor = _SpinboxEditor(
                self, from_=min, to=max, increment=step
            )
            self._editors_bag[col][_EditorType.SPINBOX] = current_editor
        else:
            current_editor = self._editors_bag[col][_EditorType.SPINBOX]
        self._editors[col] = current_editor
        self._setup_editor(col, item, current_editor)

    def inplace_custom(self, col, item, widget, stringvar=None):
        current_editor = None
        if _EditorType.CUSTOM_WIDGET not in self._editors_bag[col]:
            if stringvar is None:
                stringvar = tk.StringVar()
            current_editor = _CustomEditor(widget, textvariable=stringvar)
            self._editors_bag[col][_EditorType.CUSTOM_WIDGET] = current_editor
        else:
            current_editor = self._editors_bag[col][_EditorType.CUSTOM_WIDGET]
        self._editors[col] = current_editor
        self._setup_editor(col, item, current_editor)

    def inplace_editor(self, col, item, editor: InplaceEditor):
        current_editor = None
        if _EditorType.CUSTOM_EDITOR not in self._editors_bag[col]:
            current_editor = editor
            self._editors_bag[col][_EditorType.CUSTOM_EDITOR] = current_editor
        else:
            current_editor = self._editors_bag[col][_EditorType.CUSTOM_EDITOR]
        self._editors[col] = current_editor
        self._setup_editor(col, item, current_editor)
